"""Writer functions."""

from __future__ import annotations

import typing as ty
from pathlib import Path

import numba
import numpy as np
from koyo.system import IS_MAC_ARM
from koyo.typing import PathLike
from loguru import logger
from tqdm import tqdm

from image2image_io.utils.utilities import (
    get_dtype_for_array,
    get_flat_shape_of_image,
    get_shape_of_image,
    reshape_fortran,
)
from image2image_io.writers.merge_tiff_writer import MergeImages, MergeOmeTiffWriter
from image2image_io.writers.tiff_writer import OmeTiffWrapper, OmeTiffWriter, Transformer

if ty.TYPE_CHECKING:
    from image2image_io.readers._base_reader import BaseReader

MetadataScene = dict[str, ty.Union[str, list[ty.Union[int, str]]]]
MetadataReader = dict[int, MetadataScene]
MetadataDict = dict[Path, MetadataReader]
ExportProgress = tuple[str, int, int, int, bool]


__all__ = [
    "MergeOmeTiffWriter",
    "OmeTiffWriter",
    "czi_to_ome_tiff",
    "czis_to_ome_tiff",
    "image_to_fusion",
    "image_to_ome_tiff",
    "images_to_fusion",
    "images_to_ome_tiff",
]


def get_total_n_scenes(paths: ty.Iterable[PathLike]) -> tuple[int, list[Path]]:
    """Get total number of scenes in all files."""
    from image2image_io.readers._czi import CziSceneFile

    total_n_scenes = 0
    paths_ = []
    for path_ in paths:
        path_ = Path(path_)
        if path_.is_dir():
            for path__ in path_.glob("**/*.czi"):
                if path__.suffix == ".czi":
                    total_n_scenes += CziSceneFile.get_num_scenes(path__)
                else:
                    total_n_scenes += 1
                paths_.append(path__)
        else:
            if path_.suffix == ".czi":
                total_n_scenes += CziSceneFile.get_num_scenes(path_)
            else:
                total_n_scenes += 1
            paths_.append(path_)
    return total_n_scenes, paths_


def merge_images(
    name: str,
    paths: ty.Iterable[PathLike],
    output_dir: PathLike,
    as_uint8: bool | None = None,
    tile_size: int = 512,
    metadata: MetadataDict | None = None,
    overwrite: bool = False,
) -> None:
    """Merge multiple images to OME-TIFF."""
    from image2image_io.readers import get_simple_reader

    name = name.replace(".ome.tiff", "")
    output_dir = Path(output_dir)
    filename = output_dir / f"{name}.ome.tiff"
    if filename.exists() and not overwrite:
        logger.info(f"Skipping {filename} - already exists")
        return

    pixel_sizes = []
    channel_names = []
    channel_ids = []
    reader_names = []
    for path_ in paths:
        reader = get_simple_reader(path_, init_pyramid=False, auto_pyramid=False)
        reader_metadata: MetadataReader
        if metadata:
            reader_metadata = metadata.get(path_, None)
            if not reader_metadata:
                reader_metadata = {
                    0: {
                        "name": reader.clean_name,
                        "channel_names": reader.channel_names,
                        "channel_ids": reader.channel_ids,
                    }
                }
                logger.warning(f"Metadata not found for {path_}. Using default metadata.")
        else:
            reader_metadata = {
                0: {"name": reader.clean_name, "channel_names": reader.channel_names, "channel_ids": reader.channel_ids}
            }
            logger.trace(f"Metadata not specified for {path_}. Using default metadata.")

        scene_metadata: dict[str, list[int | str]]
        if reader_metadata:
            scene_metadata = reader_metadata.get(0, None)
            if not scene_metadata:
                scene_metadata = {"channel_ids": reader.channel_ids, "channel_names": reader.channel_names}
                logger.warning("Metadata not found for 0. Using default metadata.")
        else:
            scene_metadata = {
                "name": reader.clean_name,
                "channel_ids": reader.channel_ids,
                "channel_names": reader.channel_names,
            }
            logger.trace("Metadata not specified for 0. Using default metadata.")

        pixel_sizes.append(reader.resolution)
        reader_names.append(scene_metadata.get("name", reader.clean_name))
        channel_names.append(scene_metadata.get("channel_names", reader.channel_names))
        channel_ids.append(scene_metadata.get("channel_ids", reader.channel_ids))

    writer = MergeOmeTiffWriter(
        MergeImages(
            paths,
            pixel_sizes,
            channel_names=channel_names,
            channel_ids=channel_ids,
        )
    )
    writer.merge_write_image_by_plane(
        name,
        reader_names,
        output_dir=output_dir,
        as_uint8=as_uint8,
        tile_size=tile_size,
    )


def images_to_ome_tiff(
    paths: ty.Iterable[PathLike],
    output_dir: PathLike | None = None,
    as_uint8: bool | None = None,
    tile_size: int = 512,
    metadata: MetadataDict | None = None,
    path_to_scene: dict[Path, list[int]] | None = None,
    extras: dict[Path, dict[str, int | float | None]] | None = None,
    overwrite: bool = False,
) -> ty.Generator[ExportProgress, None, None]:
    """Convert multiple images to OME-TIFF."""
    output_dir = Path(output_dir) if output_dir else None

    # get total number of scenes
    total_n_scenes, paths = get_total_n_scenes(paths)
    current_total_scene = 0
    for _current, path_ in enumerate(tqdm(paths, desc="Converting to OME-TIFF...", total=len(paths))):
        path_ = Path(path_)
        reader_metadata = metadata.get(path_, None) if metadata else None
        reader_scenes = path_to_scene.get(path_, None) if path_to_scene else None
        if path_.suffix == ".czi":
            try:
                for key, current_file_scene, total_file_scenes, increment_by, is_exported in czi_to_ome_tiff(
                    path_,
                    output_dir=output_dir,
                    as_uint8=as_uint8,
                    tile_size=tile_size,
                    metadata=reader_metadata,
                    reader_scenes=reader_scenes,
                    overwrite=overwrite,
                ):
                    yield key, current_file_scene, total_file_scenes, current_total_scene, total_n_scenes, is_exported
                    current_total_scene += increment_by
            except (ValueError, TypeError, OSError) as err:
                logger.error(f"Could not read Czi file {path_} ({err})")
                logger.exception(err)
                continue
        else:
            try:
                for key, current_file_scene, total_file_scenes, increment_by, is_exported in image_to_ome_tiff(
                    path_,
                    output_dir=output_dir,
                    as_uint8=as_uint8,
                    tile_size=tile_size,
                    metadata=reader_metadata,
                    overwrite=overwrite,
                ):
                    yield key, current_file_scene, total_file_scenes, current_total_scene, total_n_scenes, is_exported
                    current_total_scene += increment_by
            except (ValueError, TypeError, OSError) as err:
                logger.error(f"Could not read {path_.suffix} file {path_} ({err})")
                logger.exception(err)
                continue


def image_to_ome_tiff(
    path: PathLike,
    output_dir: PathLike | None = None,
    as_uint8: bool | None = None,
    tile_size: int = 512,
    suffix: str = "",
    metadata: MetadataReader | None = None,
    transformer: Transformer | None = None,
    resolution: float | None = None,
    overwrite: bool = False,
) -> ty.Generator[ExportProgress, None, None]:
    """Convert image of any type to OME-TIFF.

    Returns
    -------
    ExportProgress
        Tuple containing the key, current scene, total scenes, current total scene, and if the scene was exported
    """
    from image2image_io.readers import get_key, get_simple_reader

    path = Path(path)

    if output_dir is None:
        output_dir = path.parent
    output_dir = Path(output_dir)
    output_dir.mkdir(exist_ok=True, parents=True)

    key = get_key(path)
    filename = path.name.replace(".ome.tiff", "").replace(path.suffix, "")
    if suffix:
        filename += suffix
    output_path = output_dir / filename
    yield key, 1, 1, 1, False

    # skip if the output file already exists
    if output_path.with_suffix(".ome.tiff").exists() and not overwrite:
        logger.info(f"Skipping {output_path} - already exists")
        yield key, 1, 1, 1, True
        return

    # read the scene
    reader = get_simple_reader(path, auto_pyramid=False, init_pyramid=False)
    if resolution:
        reader.resolution = resolution

    scene_metadata: dict[str, list[int | str]]
    if metadata:
        scene_metadata = metadata.get(path, None)
        if not scene_metadata:
            scene_metadata = {"channel_ids": reader.channel_ids, "channel_names": reader.channel_names}
            logger.warning(f"Metadata not found for {path}. Using default metadata.")
    else:
        scene_metadata = {"channel_ids": reader.channel_ids, "channel_names": reader.channel_names}
        logger.trace(f"Metadata not specified for {path}. Using default metadata.")

    if scene_metadata:
        assert "channel_ids" in scene_metadata, "Channel IDs must be specified in metadata."
        assert "channel_names" in scene_metadata, "Channel names must be specified in metadata."
    if not scene_metadata["channel_ids"]:
        yield key, 1, 1, 1, True
    else:
        write_ome_tiff_alt(
            output_path,
            reader,
            as_uint8=as_uint8,
            tile_size=tile_size,
            transformer=transformer,
            **scene_metadata,
            overwrite=overwrite,
        )
        yield key, 1, 1, 1, True


def czis_to_ome_tiff(
    paths: ty.Iterable[PathLike],
    output_dir: PathLike | None = None,
    as_uint8: bool | None = None,
    tile_size: int = 512,
    metadata: MetadataDict | None = None,
) -> ty.Generator[ExportProgress, None, None]:
    """Convert multiple CZI images to OME-TIFF."""
    # calculate true total number of scenes
    total_n_scenes, paths_ = get_total_n_scenes(paths)

    current_total_scene = 0
    for path_ in paths_:
        path_ = Path(path_)
        reader_metadata = metadata.get(path_, None) if metadata else None
        try:
            for key, current_file_scene, total_file_scenes, increment_by, is_exported in czi_to_ome_tiff(
                path_, output_dir, as_uint8, tile_size, reader_metadata
            ):
                yield key, current_file_scene, total_file_scenes, current_total_scene, total_n_scenes, is_exported
                current_total_scene += increment_by
        except (ValueError, TypeError, OSError):
            logger.error(f"Could not read Czi file {path_}")
            continue


def czi_to_ome_tiff(
    path: PathLike,
    output_dir: PathLike | None = None,
    as_uint8: bool | None = None,
    tile_size: int = 512,
    metadata: MetadataReader | None = None,
    reader_scenes: list[int] | None = None,
    scenes: list[int] | None = None,
    overwrite: bool = False,
) -> ty.Generator[ExportProgress, None, None]:
    """Convert Czi image to OME-TIFF."""
    from image2image_io.readers import get_key
    from image2image_io.readers._czi import CziSceneFile
    from image2image_io.readers.czi_reader import CziSceneImageReader

    path = Path(path)
    if output_dir is None:
        output_dir = path.parent
    output_dir = Path(output_dir)
    try:
        n = CziSceneFile.get_num_scenes(path)
        has_scenes = n != 1
    except Exception as e:
        logger.error(f"Could not read Czi file {path} - {e}")
        return
    if scenes is None:
        scenes = list(range(n))
    if reader_scenes is not None:
        scenes = [scene for scene in scenes if scene in reader_scenes]

    assert min(scenes) >= 0, "Scene index must be greater than or equal to 0."
    assert max(scenes) <= n, "Scene index must be less than the total number of scenes in the file."
    scene_key = get_key(path, scenes[0] if has_scenes else None)
    yield scene_key, 0, len(scenes), 0, False

    if metadata is None:
        metadata = {}
        logger.trace("Metadata not specified. Setting to empty.")

    # iterate over each scene in the czi file
    for scene_index in scenes:
        scene_key = get_key(path, scene_index if has_scenes else None)
        logger.debug(f"Converting scene {scene_index + 1}/{n} from {path}...")
        filename = path.name.replace(".czi", "") + (f"_scene={scene_index:02d}" if n > 1 else "")
        output_path = output_dir / filename
        # skip if the output file already exists
        if output_path.with_suffix(".ome.tiff").exists() and not overwrite:
            logger.info(f"Skipping {output_path} - already exists")
            yield scene_key, scene_index + 1, n, 1, True
            continue

        # read the scene
        reader = CziSceneImageReader(path, scene_index=scene_index, auto_pyramid=False, init_pyramid=False)
        path = reader.path

        scene_metadata: dict[str, list[int | str]]
        if metadata:
            scene_metadata = metadata.get(scene_index, None)
            if not scene_metadata:
                scene_metadata = {"channel_ids": reader.channel_ids, "channel_names": reader.channel_names}
                logger.warning(
                    f"Metadata not found for {path} {scene_index}. Using default metadata. "
                    f"Available keys: {metadata.keys()}"
                )
        else:
            scene_metadata = {"channel_ids": reader.channel_ids, "channel_names": reader.channel_names}
            logger.trace(f"Metadata not specified for {path} {scene_index}. Using default metadata.")

        if scene_metadata:
            assert "channel_ids" in scene_metadata, "Channel IDs must be specified in metadata."
            assert "channel_names" in scene_metadata, "Channel names must be specified in metadata."

        # skip if there are no channel IDs
        if not scene_metadata["channel_ids"]:
            yield scene_key, scene_index + 1, n, 1, True
        else:
            write_ome_tiff_alt(output_path, reader, as_uint8=as_uint8, tile_size=tile_size, **scene_metadata)
            yield scene_key, scene_index + 1, n, 1, True


def write_ome_tiff_from_array(
    path: PathLike,
    reader: BaseReader | None,
    array: np.ndarray,
    resolution: float | None = None,
    channel_names: list[str] | None = None,
    compression: str | None = "default",
    tile_size: int = 512,
    as_uint8: bool | None = None,
    ome_name: str | None = None,
) -> Path:
    """Write OME-TIFF by also specifying an array."""
    from image2image_io.readers.array_reader import ArrayImageReader
    from image2image_io.writers.tiff_writer import OmeTiffWriter

    if array.ndim == 2:
        array = array[np.newaxis, ...]

    if reader:
        resolution = resolution or reader.resolution
        channel_names = channel_names or reader.channel_names
    resolution = resolution or 1.0

    array_reader = ArrayImageReader("", array, resolution=resolution, channel_names=channel_names)

    path = Path(path)
    filename = path.name.replace(".ome.tiff", "")
    writer = OmeTiffWriter(array_reader)
    output_path = writer.write_image_by_plane(
        filename,
        path.parent,
        write_pyramid=True,
        compression=compression,
        tile_size=tile_size,
        as_uint8=as_uint8,
        ome_name=ome_name,
    )
    return output_path


def write_ome_tiff_alt(
    path: PathLike,
    reader: BaseReader,
    as_uint8: bool | None = None,
    tile_size: int = 512,
    channel_ids: list[int] | None = None,
    channel_names: list[str] | None = None,
    transformer: Transformer | None = None,
    overwrite: bool = False,
    write: bool = True,
    ome_name: str | None = None,
) -> Path:
    """Write OME-TIFF."""
    from image2image_io.writers.tiff_writer import OmeTiffWriter

    path = Path(path)
    filename = path.name.replace(".ome.tiff", "")
    writer = OmeTiffWriter(reader, transformer=transformer)
    if write:
        output_path = writer.write_image_by_plane(
            filename,
            path.parent,
            write_pyramid=tile_size > 0,
            as_uint8=as_uint8,
            channel_ids=channel_ids,
            channel_names=channel_names,
            tile_size=tile_size,
            overwrite=overwrite,
            ome_name=ome_name,
        )
        return output_path
    return writer


def images_to_fusion(
    paths: ty.Iterable[PathLike],
    output_dir: PathLike | None = None,
    metadata: MetadataDict | None = None,
) -> ty.Generator[tuple[str, int, int, str], None, None]:
    """Convert multiple images to Fusion."""
    for path_ in paths:
        path_ = Path(path_)
        reader_metadata = metadata.get(path_, None) if metadata else None
        if reader_metadata:
            reader_metadata = reader_metadata.get(0, None)
        yield from image_to_fusion(path_, output_dir, reader_metadata)


def image_to_fusion(
    path: PathLike,
    output_dir: PathLike | None = None,
    metadata: MetadataReader | None = None,
) -> ty.Generator[tuple[str, int, int, str], None, None]:
    """Convert image of any type to Fusion format."""
    from image2image_io.readers import get_reader

    path = Path(path)
    path, readers = get_reader(path)
    for reader in readers.values():
        key = reader.key
        if output_dir is None:
            output_dir = path.parent
        output_dir = Path(output_dir)
        filename = output_dir / path.stem.replace(".ome", "")
        yield from write_reader_to_txt(reader, filename, metadata=metadata, key=key)
        logger.trace(f"Exported Fusion CSV for {path}")
        if not filename.with_suffix(".xml").exists():
            write_reader_to_xml(reader, filename)
            logger.trace(f"Exported XML metadata for {path}")
        del reader


def write_reader_to_xml(reader: BaseReader, filename: PathLike) -> None:
    """Get filename."""
    from xml.dom.minidom import parseString

    from dicttoxml import dicttoxml

    _, _, image_shape = get_shape_of_image(reader.shape)
    shape = get_flat_shape_of_image(reader.shape)

    meta = {
        "modality": "microscopy",
        "data_label": reader.path.stem,
        "nr_spatial_dims": 2,
        "spatial_grid_size": f"{image_shape[0]} {image_shape[1]}",
        "nr_spatial_grid_elems": image_shape[0] * image_shape[1],
        "spatial_resolution_um": reader.resolution,
        "nr_obs": shape[0],
        "nr_vars": shape[1],
    }
    xml = dicttoxml(meta, custom_root="data_source", attr_type=False)

    # filename = xml_output_path.parent / (reader.path.stem + ".xml")
    filename = Path(filename).with_suffix(".xml")
    with open(filename, "w") as f:
        f.write(parseString(xml).toprettyxml())
    logger.debug(f"Saved XML file to {filename}")


def write_reader_to_txt(
    reader: BaseReader,
    filename: PathLike,
    metadata: MetadataReader | None = None,
    update_freq: int = 100_000,
    key: str = "",
) -> ty.Generator[tuple[str, int, int, str], None, None]:
    """Write image data to text."""
    filename = Path(filename).with_suffix(".txt")
    if filename.exists():
        yield key, 0, 0, ""

    # convert to appropriate format
    array, shape = reader.flat_array()
    array = _insert_indices(array, shape)

    if reader.is_rgb and reader.dtype == np.uint8:
        channel_names = ["Red (R)", "Green (G)", "Blue (B)"]
        yield from write_rgb_to_txt(filename, array, channel_names, update_freq=update_freq, key=key)
    elif np.issubdtype(reader.dtype, np.integer):
        yield from write_int_to_txt(filename, array, reader.channel_names, update_freq=update_freq, key=key)
    else:
        yield from write_float_to_txt(filename, array, reader.channel_names, update_freq=update_freq, key=key)


def float_format_to_row(row: np.ndarray) -> str:
    """Format string to row."""
    return ",".join([",".join([str(int(v)) for v in row[0:2]]), ",".join([f"{v:.2f}" for v in row[2:]]), "\n"])


@numba.njit(nogil=True, cache=True)  # type: ignore[misc]
def int_format_to_row(row: np.ndarray) -> str:
    """Format string to row."""
    return ",".join([str(v) for v in row]) + ",\n"


def _insert_indices(array: np.ndarray, shape: tuple[int, int]) -> np.ndarray:
    n = array.shape[1]
    y, x = np.indices(shape)
    y = y.ravel(order="F")
    x = x.ravel(order="F")
    dtype = get_dtype_for_array(array)
    res = np.zeros((x.size, n + 2), dtype=dtype)
    res[:, 0] = y + 1
    res[:, 1] = x + 1
    res[:, 2:] = reshape_fortran(array, (-1, n))
    return res


def write_rgb_to_txt(
    path: PathLike, array: np.ndarray, channel_names: list[str], update_freq: int = 100_000, key: str = ""
) -> ty.Generator[tuple[str, int, int, str], None, None]:
    """Write IMS data to text."""
    columns = ["row", "col", *channel_names]
    fmt_func = int_format_to_row.py_func if IS_MAC_ARM else int_format_to_row
    yield from _write_txt(path, columns, array, fmt_func, update_freq=update_freq, key=key)


def write_int_to_txt(
    path: PathLike, array: np.ndarray, channel_names: list[str], update_freq: int = 100_000, key: str = ""
) -> ty.Generator[tuple[str, int, int, str], None, None]:
    """Write IMS data to text."""
    columns = ["row", "col", *channel_names]
    fmt_func = int_format_to_row.py_func if IS_MAC_ARM else int_format_to_row
    yield from _write_txt(path, columns, array, fmt_func, update_freq=update_freq, key=key)


def write_float_to_txt(
    path: PathLike, array: np.ndarray, channel_names: list[str], update_freq: int = 100_000, key: str = ""
) -> ty.Generator[tuple[str, int, int, str], None, None]:
    """Write IMS data to text."""
    columns = ["row", "col", *channel_names]
    yield from _write_txt(path, columns, array, float_format_to_row, update_freq=update_freq, key=key)


def _write_txt(
    path: PathLike,
    columns: list[str],
    array: np.ndarray,
    str_func: ty.Callable,
    update_freq: int = 100_000,
    key: str = "",
) -> ty.Generator[tuple[str, int, int, str], None, None]:
    """Write data to csv file."""
    from tqdm import tqdm

    path = Path(path)
    assert path.suffix == ".txt", "Path must have .txt extension."

    n = array.shape[0]
    logger.debug(
        f"Exporting array with {array.shape[0]:,} observations, {array.shape[1]:,} features and {array.dtype} data"
        f" type to '{path}'"
    )
    with open(path, "w", newline="\n", encoding="cp1252") as f:
        f.write(",".join(columns) + ",\n")
        with tqdm(array, total=n, mininterval=1, desc="Exporting to CSV...") as pbar:
            for i, row in enumerate(pbar):
                f.write(str_func(row))
                if i % update_freq == 0 and i != 0:
                    d = pbar.format_dict
                    if d["rate"]:
                        eta = pbar.format_interval((d["total"] - d["n"]) / d["rate"])
                    else:
                        eta = ""
                    yield key, i, n, eta
        yield key, n, n, ""



import click


# noinspection PyUnusedLocal
def arg_split_bbox(ctx, param, value):
    """Split arguments."""
    if value is None:
        return None
    args = [int(arg.strip()) for arg in value.split(",")]
    assert len(args) == 4, "Bounding box must have 4 values"
    return args


ALLOW_EXTRA_ARGS = {"help_option_names": ["-h", "--help"], "ignore_unknown_options": True, "allow_extra_args": True}
overwrite_ = click.option(
    "-W",
    "--overwrite",
    help="Overwrite existing data.",
    is_flag=True,
    default=False,
    show_default=True,
)
as_uint8_ = click.option(
    "-u/-U",
    "--as_uint8/--no_as_uint8",
    help="Downcast the image data format to uint8 which will substantially reduce the size of the files (unless it's"
    " already in uint8...).",
    is_flag=True,
    default=None,
    show_default=True,
)
fmt_ = click.option(
    "-f",
    "--fmt",
    help="Output format.",
    type=click.Choice(["ome-tiff"], case_sensitive=False),
    default="ome-tiff",
    show_default=True,
    required=False,
)



"""Utilities."""

from __future__ import annotations

import typing as ty
from collections import defaultdict
from pathlib import Path
from xml.etree import ElementTree as ET

import numpy as np
from koyo.typing import PathLike

if ty.TYPE_CHECKING:
    from skimage.transform import ProjectiveTransform


def resize(array: np.ndarray, shape: tuple[int, int]) -> np.ndarray:
    """Resize image which could be grayscale, RGB or multi-channel."""
    import cv2

    _, channel_axis, _ = get_shape_of_image(array)
    if channel_axis is None:
        array = cv2.resize(array, shape[::-1], interpolation=cv2.INTER_LINEAR)
    else:
        array = cv2.resize(array, shape[::-1], interpolation=cv2.INTER_LINEAR)
    return array


def write_thumbnail(path: PathLike, output_dir: PathLike, with_title: bool, first_only: bool = False) -> None:
    """Write thumbnail."""
    import matplotlib

    from image2image_io.readers import get_simple_reader

    matplotlib.use("agg")

    output_dir = Path(output_dir)
    output_dir.mkdir(parents=True, exist_ok=True)
    reader = get_simple_reader(path, auto_pyramid=False)
    if reader.is_rgb:
        image = reader.pyramid[-1]
        filename = output_dir / f"{reader.name}_rgb_thumbnail.jpg"
        make_thumbnail(filename, image, f"{reader.name}", with_title)
    else:
        for channel_id, channel_name in enumerate(reader._channel_names):
            filename = output_dir / f"{reader.name}_index={channel_id}_name={channel_name}_thumbnail.jpg"
            image = reader.get_channel(channel_id, -1)
            make_thumbnail(filename, image, f"{reader.name}\n{channel_id} / {channel_name}", with_title)
            if first_only:
                break
    del reader


def make_thumbnail(filename: Path, image: np.ndarray, title: str, with_title: bool):
    """Create thumbnail of an image."""
    import matplotlib.pyplot as plt

    fig, ax = plt.subplots(figsize=(12, 12))
    if image.ndim == 2:
        vmax = np.percentile(np.ravel(image), 99.0)
        ax.imshow(image, cmap="turbo", aspect="equal", vmax=vmax)
    else:
        ax.imshow(image, aspect="equal")
    if with_title:
        ax.set_title(title, fontsize=14)
    ax.axis("off")
    plt.savefig(filename, bbox_inches="tight")
    plt.close()


def xmlstr_to_dict(xmlstr: str) -> dict:
    """Convert xml string to dict."""
    return etree_to_dict(ET.fromstring(xmlstr))


def etree_to_dict(t: ET) -> dict:
    d = {t.tag: {} if t.attrib else None}
    children = list(t)
    if children:
        dd = defaultdict(list)
        for dc in map(etree_to_dict, children):
            for k, v in dc.items():
                dd[k].append(v)
        d = {t.tag: {k: v[0] if len(v) == 1 else v for k, v in dd.items()}}
    if t.attrib:
        d[t.tag].update(("@" + k, v) for k, v in t.attrib.items())
    if t.text:
        text = t.text.strip()
        if children or t.attrib:
            if text:
                d[t.tag]["#text"] = text
        else:
            d[t.tag] = text
    return d


def format_merge_channel_names(
    channel_names: list[list[str]] | None, n_ch: int, channel_ids: list[list[int]] | None
) -> list[str]:
    """Format channel names and ensure number of channel names matches number of channels or default to C1, C2, C3, etc.

    Parameters
    ----------
    channel_names:list
        list of str that are channel names
    n_ch: int
        number of channels detected in image
    channel_ids: list[list[int]] | None
        list of channel ids

    Returns
    -------
    channel_names:
        list of str that are formatted
    """
    if channel_names is None:
        channel_names = [f"C{idx}" for idx in range(n_ch)]
    if channel_ids is not None:
        channel_names_ = []
        if len(channel_ids) != len(channel_names):
            raise ValueError("The number of channel ids does not match the number of channel names.")
        for index, channel_ids_ in enumerate(channel_ids):
            channel_names_.extend([channel_names[index][idx_] for idx_ in channel_ids_])
        channel_names = channel_names_
    else:
        if isinstance(channel_names, list) and isinstance(channel_names[0], list):
            channel_names = [name for names in channel_names for name in names]
        if n_ch != len(channel_names):
            raise ValueError("The number of channels does not match the number of channel names.")
    return channel_names


def format_mz(mz: float) -> str:
    """Format m/z value."""
    return f"m/z {mz:.3f}"


def guess_rgb(shape: tuple[int, ...]) -> bool:
    """Guess if the passed shape comes from rgb data.

    If last dim is 3 or 4 assume the data is rgb, including rgba.

    Parameters
    ----------
    shape : list of int
        Shape of the data that should be checked.

    Returns
    -------
    bool
        If data is rgb or not.
    """
    if hasattr(shape, "shape"):
        shape = shape.shape
    ndim = len(shape)
    last_dim = shape[-1]
    rgb = False
    if ndim > 2 and last_dim < 5:
        rgb = True
    return rgb


def get_shape_of_image(array_or_shape: np.ndarray | tuple[int, ...]) -> tuple[int, int | None, tuple[int, int]]:
    """Return shape of an image."""
    if isinstance(array_or_shape, tuple):
        ndim = len(array_or_shape)
        shape = array_or_shape
    else:
        ndim = array_or_shape.ndim
        shape = array_or_shape.shape

    shape = list(shape)  # type: ignore[assignment]
    if ndim == 2:
        n_channels = 1
        channel_axis = None
    elif ndim == 3:
        if shape[2] in [3, 4]:  # rgb or rgba
            channel_axis = 2
            n_channels = shape[2]
        else:
            channel_axis = 0
            n_channels = shape[0]
        # channel_axis = int(np.argmin(shape))
        # n_channels = int(shape[channel_axis])
        shape.pop(channel_axis)
    else:
        n_channels = 1
        channel_axis = None
    return n_channels, channel_axis, tuple(shape)


def get_flat_shape_of_image(array_or_shape: np.ndarray | tuple[int, ...]) -> tuple[int, int]:
    """Return shape of an image."""
    n_channels, _, shape = get_shape_of_image(array_or_shape)
    n_px = int(np.prod(shape))
    return n_channels, n_px


def compute_transform(src: np.ndarray, dst: np.ndarray, transform_type: str = "affine") -> ProjectiveTransform:
    """Compute transform."""
    from skimage.transform import estimate_transform

    if len(dst) != len(src):
        raise ValueError(f"The number of fixed and moving points is not equal. (moving={len(dst)}; fixed={len(src)})")
    return estimate_transform(transform_type, src, dst)


def get_dtype_for_array(array: np.ndarray) -> np.dtype:
    """Return smallest possible data type for shape."""
    n = array.shape[1]
    if np.issubdtype(array.dtype, np.integer):
        if n < np.iinfo(np.uint8).max:
            return np.uint8
        elif n < np.iinfo(np.uint16).max:
            return np.uint16
        elif n < np.iinfo(np.uint32).max:
            return np.uint32
        elif n < np.iinfo(np.uint64).max:
            return np.uint64
    else:
        if n < np.finfo(np.float32).max:
            return np.float32
        elif n < np.finfo(np.float64).max:
            return np.float64


def reshape_fortran(x: np.ndarray, shape: tuple[int, int]) -> np.ndarray:
    """Reshape data to Fortran (MATLAB) ordering."""
    return x.T.reshape(shape[::-1]).T


def reshape(x: np.ndarray, y: np.ndarray, array: np.ndarray, fill_value: float = np.nan) -> np.ndarray:
    """Reshape array."""
    xmin, xmax = np.min(x), np.max(x)
    ymin, ymax = np.min(y), np.max(y)
    shape = (ymax - ymin + 1, xmax - xmin + 1)
    dtype = np.float32 if np.isnan(fill_value) else array.dtype
    new_array = np.full(shape, fill_value=fill_value, dtype=dtype)
    new_array[y - ymin, x - xmin] = array
    return new_array


def reshape_batch(x: np.ndarray, y: np.ndarray, array: np.ndarray, fill_value: float = np.nan) -> np.ndarray:
    """Batch reshaping of images."""
    if array.ndim != 2:
        raise ValueError("Expected 2-D array.")
    xmin, xmax = np.min(x), np.max(x)
    ymin, ymax = np.min(y), np.max(y)
    y = y - ymin
    x = x - xmin
    n = array.shape[1]
    shape = (n, ymax - ymin + 1, xmax - xmin + 1)
    dtype = np.float32 if np.isnan(fill_value) else array.dtype
    im = np.full(shape, fill_value=fill_value, dtype=dtype)
    for i in range(n):
        im[i, y, x] = array[:, i]
    return im


def get_yx_coordinates_from_shape(shape: tuple[int, int]) -> tuple[np.ndarray, np.ndarray]:
    """Get coordinates from image shape."""
    _y, _x = np.indices(shape)
    yx_coordinates = np.c_[np.ravel(_y), np.ravel(_x)]
    return yx_coordinates[:, 0], yx_coordinates[:, 1]



"""Merge command."""

from __future__ import annotations

import typing as ty

import click
from koyo.click import Parameter, arg_parse_framelist_multi, cli_parse_paths_sort, print_parameters
from koyo.timer import MeasureTimer
from loguru import logger

from image2image_io.cli._common import as_uint8_, fmt_, overwrite_
from image2image_io.enums import WriterMode


@overwrite_
@as_uint8_
@fmt_
@click.option(
    "-C",
    "--channel_ids",
    type=click.STRING,
    default=None,
    help="Specify channel ids in the format: 1,2,4-6. You can provide multiple. If you are providing any, make sure to"
    " provide one for each file you are trying to merge.",
    callback=arg_parse_framelist_multi,
    show_default=True,
    multiple=True,
    required=False,
)
@click.option(
    "-o",
    "--output_dir",
    help="Path to images to be merged.",
    type=click.Path(file_okay=False, dir_okay=True, resolve_path=True),
    show_default=True,
    required=True,
)
@click.option(
    "-p",
    "--path",
    help="Path to images to be merged.",
    type=click.UNPROCESSED,
    show_default=True,
    required=True,
    multiple=True,
    callback=cli_parse_paths_sort,
)
@click.option(
    "-n",
    "--name",
    help="Name of the merged image.",
    type=click.STRING,
    show_default=True,
    required=True,
)
@click.command("merge")
def merge(
    name: str,
    path: ty.Sequence[str],
    output_dir: str,
    channel_ids: ty.Sequence[tuple] | None,
    fmt: WriterMode,
    as_uint8: bool | None,
    overwrite: bool,
) -> None:
    """Export images."""
    merge_runner(name, path, output_dir, channel_ids, fmt, as_uint8, overwrite)


def merge_runner(
    name: str,
    paths: ty.Sequence[str],
    output_dir: str,
    channel_ids: ty.Sequence[tuple] | None,
    fmt: WriterMode = "ome-tiff",
    as_uint8: bool | None = False,
    overwrite: bool = False,
) -> None:
    """Register images."""
    from image2image_io.workflows.merge import merge as merge_images

    print_parameters(
        Parameter("Name", "-n/--name", name),
        Parameter("Image paths", "-p/--path", paths),
        Parameter("Output directory", "-o/--output_dir", output_dir),
        Parameter("Channel ids", "-C/--channel_ids", channel_ids),
        Parameter("Output format", "-f/--fmt", fmt),
        Parameter("Write images as uint8", "--as_uint8/--no_as_uint8", as_uint8),
        Parameter("Overwrite", "-W/--overwrite", overwrite),
    )

    if channel_ids:
        if len(channel_ids) != len(paths):
            raise ValueError("Number of channel ids must match number of images.")

    with MeasureTimer() as timer:
        merge_images(name, list(paths), output_dir, fmt, as_uint8, channel_ids=channel_ids, overwrite=overwrite)
    logger.info(f"Finished processing project in {timer()}.")



"""OME-TIFF writer for MergeRegImage class."""

from __future__ import annotations

import typing as ty
from pathlib import Path

import cv2
import numpy as np
import SimpleITK as sitk
from koyo.decorators import retry
from koyo.timer import MeasureTimer
from loguru import logger
from tifffile import TiffWriter
from tqdm import tqdm

from image2image_io.enums import SITK_TO_NP_DTYPE
from image2image_io.models.merge import MergeImages
from image2image_io.readers import BaseReader
from image2image_io.readers.utilities import get_pyramid_info, prepare_ome_xml_str
from image2image_io.utils.utilities import format_merge_channel_names


class Transformer(ty.Protocol):
    """Transformer protocol to transform image."""

    output_size: tuple[int, int]
    output_spacing: tuple[float, float]

    def __call__(self, image: sitk.Image) -> sitk.Image:
        """Transform image."""
        ...


class MergeOmeTiffWriter:
    """Class for writing multiple images into a single OME-TIFF.

    Attributes
    ----------
    x_size: int
        Size of the merged image after transformation in x
    y_size: int
        Size of the merged image after transformation in y
    tile_size: int
        Size of tiles to be written
    pyr_levels: list of tuples of int:
        Size of down-sampled images in pyramid
    n_pyr_levels: int
        Number of downsamples in pyramid
    PhysicalSizeY: float
        physical size of image in micron for OME-TIFF in Y
    PhysicalSizeX: float
        physical size of image in micron for OME-TIFF in X
    subifds: int
        Number of sub-resolutions for pyramidal OME-TIFF
    compression: str
        tifffile string to pass to compression argument, defaults to "deflate" for minisblack
        and "jpeg" for RGB type images
    """

    x_size: int | None = None
    y_size: int | None = None
    tile_size: int = 512
    pyr_levels: list[tuple[int, int]]
    n_pyr_levels: int
    PhysicalSizeY: int | float
    PhysicalSizeX: int | float
    subifds: int | None = None
    compression: str = "deflate"

    def __init__(
        self,
        merge: MergeImages,
        transformers: list[Transformer | None] | None = None,
        crop_mask: np.ndarray | None = None,
        crop_bbox: tuple[int, int, int, int] | None = None,
    ):
        """
        Class for writing multiple images with and without transforms to a single OME-TIFF.

        Parameters
        ----------
        merge: MergeImages
            MergeRegImage to be transformed
        transformers: List of RegTransformSeq or None
            Registration transformation sequences for each wsireg image to be merged
        crop_mask : np.ndarray or None
            Crop mask to apply to images before writing.
        crop_bbox : tuple of ints or None
            Bounding box to crop images to before writing. Values should be x, y, width, height.
        """
        self.merge = merge
        self.transformers = transformers
        self.crop_mask = crop_mask
        self.crop_bbox = crop_bbox
        if self.crop_mask is not None and self.crop_bbox is not None:
            raise ValueError("Cannot supply both crop_mask and crop_bbox")

    @staticmethod
    def _check_bbox(
        crop_bbox: tuple[int, int, int, int] | None, transformer: Transformer | None, reader: BaseReader
    ) -> tuple[int, int, int, int] | None:
        """Check bbox."""
        if crop_bbox is None:
            return None
        x, y, width, height = crop_bbox
        if x < 0:
            x = 0
        if y < 0:
            y = 0
        if transformer is not None:
            image_shape = transformer.output_size
        else:
            image_shape = reader.image_shape
        if x + width > image_shape[1]:
            width = image_shape[1] - x
        if y + height > image_shape[0]:
            height = image_shape[0] - y
        return x, y, width, height

    def _check_transforms_and_readers(self) -> None:
        """Make sure incoming data is kosher in dimensions."""
        transformations = self.transformers
        if transformations is None:
            transformations = [None for _ in range(len(self.merge.readers))]
        if len(transformations) != len(self.merge.readers):
            raise ValueError("The number of transforms does not match number of images")

    def _create_channel_names(self, reader_names: list[str]) -> None:
        """Create channel names for merge data."""

        def _prepare_channel_names(reader_name: str, channel_names_: list[str]) -> list[str]:
            return [f"{channel_name} - {reader_name}" for channel_name in channel_names_]

        all_channel_names = []
        for channel_names in self.merge.channel_names:
            all_channel_names.extend(channel_names)

        # if len(set(all_channel_names)) != self.merge.n_channels:
        self.merge.channel_names = [
            _prepare_channel_names(name, channel_names)
            for name, channel_names in zip(reader_names, self.merge.channel_names)
        ]

    def _check_transforms_sizes_and_resolutions(self) -> None:
        """Check that all transforms as currently loaded output to the same size/resolution."""
        out_size = []
        out_spacing = []
        registered_transforms = self.transformers
        if registered_transforms is None:
            registered_transforms = [None for _ in range(len(self.merge.readers))]
        for reader, transform in zip(self.merge.readers, registered_transforms):
            if transform:
                out_size.append(transform.output_size)
                out_spacing.append(transform.output_spacing)
            else:
                out_size.append(reader.image_shape)
                out_spacing.append((reader.resolution, reader.resolution))

        if not all(out_spacing):
            raise ValueError("All transforms output spacings and untransformed image spacings must match")
        if not all(out_size):
            raise ValueError("All transforms output sizes and untransformed image sizes must match")

    def _check_channel_ids(self, channel_ids: list[list[int] | None] | None) -> list[list[int]]:
        """Check channel ids."""
        if channel_ids is None or not channel_ids:
            channel_ids = [reader.channel_ids for reader in self.merge.readers]
        if len(channel_ids) != len(self.merge.readers):
            raise ValueError(
                f"The number of channel_ids does not match number of images. Expected {len(self.merge.readers)} but"
                f" got {len(channel_ids)}"
            )
        channel_ids_ret = []
        for reader_index, channel_ids_ in enumerate(channel_ids):
            if channel_ids_ is None:
                channel_ids_ = list(range(0, self.merge.readers[reader_index].n_channels))
            if channel_ids_ is not None:
                if len(channel_ids_) == 0:
                    raise ValueError("Channel ids cannot be empty")
                elif max(channel_ids_) > self.merge.readers[reader_index].n_channels - 1:
                    raise ValueError("Channel ids cannot be larger than the number of channels in the image")
                channel_ids_ret.append(channel_ids_)
        return channel_ids_ret

    def _prepare_image_info(
        self,
        reader: BaseReader,
        image_name: str,
        dtype: np.dtype,
        channel_ids: list[list[int]] | None = None,
        transformer: Transformer | None = None,
        write_pyramid: bool = True,
        tile_size: int = 512,
        compression: str = "default",
    ) -> None:
        """Prepare OME-XML and other data needed for saving."""
        if transformer:
            self.x_size, self.y_size = transformer.output_size[::-1]
        elif self.crop_mask is not None:
            self.y_size, self.x_size = self.crop_mask.shape
        else:
            self.y_size, self.x_size = reader.image_shape

        self.crop_bbox = self._check_bbox(self.crop_bbox, transformer, reader)

        if self.crop_bbox is not None:
            _, _, self.x_size, self.y_size = self.crop_bbox

        self.tile_size = tile_size
        # protect against too large tile size
        while self.y_size / self.tile_size <= 1 or self.x_size / self.tile_size <= 1:
            self.tile_size = self.tile_size // 2

        if self.tile_size:
            self.pyr_levels, _ = get_pyramid_info(self.y_size, self.x_size, reader.n_channels, self.tile_size)
        else:
            self.pyr_levels = [(self.y_size, self.x_size)]
        self.n_pyr_levels = len(self.pyr_levels)

        if transformer:
            self.PhysicalSizeX, self.PhysicalSizeY = transformer.output_spacing
        else:
            self.PhysicalSizeX = self.PhysicalSizeY = reader.resolution

        channel_names = format_merge_channel_names(self.merge.channel_names, self.merge.n_channels, channel_ids)
        logger.trace(f"Exporting: {channel_names} for {channel_ids}")
        n_channels = len(channel_names)
        self.omexml = prepare_ome_xml_str(
            self.y_size,
            self.x_size,
            n_channels,
            dtype,
            False,
            PhysicalSizeX=self.PhysicalSizeX,
            PhysicalSizeY=self.PhysicalSizeY,
            PhysicalSizeXUnit="µm",
            PhysicalSizeYUnit="µm",
            Name=image_name,
            Channel={"Name": channel_names},
        )

        self.subifds = self.n_pyr_levels - 1 if write_pyramid is True else None

        if compression == "default":
            logger.trace("using default compression")
            self.compression = "deflate"
        else:
            self.compression = compression

    def _get_merge_dtype(self, as_uint8: bool = False) -> tuple[int, np.dtype[ty.Any]]:
        """Determine the data type for merger - will default to the largest dtype.

        If one image is np.uint8 and another np.uint16, the image at np.uint8 will be cast to np.uint16.
        """
        dtype_max_size = [
            (np.finfo(r.dtype).max if np.issubdtype(r.dtype, np.floating) else np.iinfo(r.dtype).max)
            for r in self.merge.readers
        ]

        merge_dtype_np: np.dtype[ty.Any] = self.merge.readers[np.argmax(dtype_max_size)].dtype  # type: ignore[arg-type]
        merge_dtype_np = np.uint8 if as_uint8 else merge_dtype_np  # type: ignore[assignment]
        merge_dtype_sitk = 0
        for k, v in SITK_TO_NP_DTYPE.items():
            if k < 12:
                if v == merge_dtype_np:
                    merge_dtype_sitk = k
        return merge_dtype_sitk, merge_dtype_np

    def merge_write_image_by_plane(
        self,
        name: str,
        reader_names: list[str],
        output_dir: Path | str = "",
        write_pyramid: bool = True,
        tile_size: int = 512,
        compression: str = "default",
        as_uint8: bool | None = False,
        channel_ids: list[list[int | tuple[int, ...]] | None] | None = None,
    ) -> Path:
        """Write merged OME-TIFF image plane-by-plane to disk.

         RGB images will be de-interleaved with RGB channels written as separate planes.

        Parameters
        ----------
        name: str
             Name to be written WITHOUT the extension, for example, if image_name = "cool_image" the file
             would be "cool_image.ome.tiff"
        reader_names: list of str
            Names added before each channel of a given image to distinguish it.
        output_dir: Path or str
             Directory where the image will be saved
        write_pyramid: bool
             Whether to write the OME-TIFF with sub-resolutions or not
        tile_size: int
             What size to write OME-TIFF tiles to disk
        compression: str
             tifffile string to pass to compression argument, defaults to "deflate" for minisblack
             and "jpeg" for RGB type images
        as_uint8: bool
            Whether to save the image in the 0-255 intensity range, substantially reducing file size at the cost of some
            intensity information.
        channel_ids: list of list
            Channel indices to write to OME-TIFF, if None, all channels are written

        Returns
        -------
         output_file_name: str
             File path to the written OME-TIFF
        """
        merge_dtype_sitk, merge_dtype_np = self._get_merge_dtype(as_uint8=as_uint8)

        channel_ids_fixed: list[list[int]] = self._check_channel_ids(channel_ids)
        self._check_transforms_and_readers()
        self._create_channel_names(reader_names)
        self._check_transforms_sizes_and_resolutions()
        assert isinstance(channel_ids_fixed, list), "channel_ids must be a list of lists"

        # make sure user did not provide filename with OME-TIFF
        name = name.replace(".ome", "").replace(".tiff", "").replace(".tif", "")
        output_file_name = Path(output_dir) / f"{name}.ome.tiff"
        tmp_output_file_name = Path(output_dir) / f"{name}.ome.tiff.tmp"
        logger.info(f"Saving to '{output_file_name}'")
        transformer = self.transformers[0] if self.transformers else None
        self._prepare_image_info(
            self.merge.readers[0],
            name,
            merge_dtype_np,
            channel_ids=channel_ids_fixed,
            transformer=transformer,
            write_pyramid=write_pyramid,
            tile_size=tile_size,
            compression=compression,
        )

        if as_uint8:
            logger.trace("Writing image data in 0-255 range as uint8")
        logger.info(f"Writing channels: {channel_ids_fixed}")
        logger.info(f"Writing channel names: {self.merge.channel_names}")

        options = {
            "tile": (tile_size, tile_size),
            "compression": self.compression,
            "photometric": "minisblack",
            "metadata": None,
        }
        logger.trace(f"TIFF options: {options}")
        logger.trace(f"Pyramid levels: {self.pyr_levels} ({self.n_pyr_levels})")

        ome_set = False
        description = None
        with TiffWriter(tmp_output_file_name, bigtiff=True) as tif, MeasureTimer() as main_timer:
            for reader_index, reader in enumerate(tqdm(self.merge.readers, desc="Writing modality...")):
                channel_ids_ = channel_ids_fixed[reader_index]
                if channel_ids_ is None:
                    channel_ids_ = list(range(0, reader.n_channels))
                for channel_index in tqdm(
                    channel_ids_, leave=False, desc=f"Exporting images for '{reader_names[reader_index]}'"
                ):
                    channel_name = self.merge.channel_names[reader_index][channel_index]
                    if channel_index not in channel_ids_:
                        logger.trace(f"Skipped channel: {channel_name} ({channel_index})")
                        continue

                    # retrieve channel data
                    # if channel_index is int, then it's simple, otherwise, let's get maximum intensity projection
                    if isinstance(channel_index, int):
                        image: sitk.Image = np.squeeze(reader.get_channel(channel_index))  # type: ignore[assignment]
                    else:
                        images: list[sitk.Image] = [np.squeeze(reader.get_channel(i)) for i in channel_index]
                        images = np.dstack(images)  # type: ignore[assignment]
                        image = np.max(images, axis=2)  # type: ignore[assignment]
                    image: sitk.Image = sitk.GetImageFromArray(image)  # type: ignore[no-redef]
                    image.SetSpacing((reader.resolution, reader.resolution))  # type: ignore[attr-defined]

                    # transform
                    if self.transformers and callable(self.transformers[reader_index]):
                        with MeasureTimer() as timer:
                            image = self.transformers[reader_index](image)  # type: ignore[assignment,arg-type,misc]
                        logger.trace(
                            f"Transformed image shape: {image.GetSize()[::-1]} in {timer()}",  # type: ignore[attr-defined]
                        )

                    # change dtype
                    if as_uint8:
                        image = sitk.RescaleIntensity(image, 0, 255)  # type: ignore[no-untyped-call]
                        image = sitk.Cast(image, sitk.sitkUInt8)  # type: ignore[no-untyped-call]
                    elif image.GetPixelIDValue() != merge_dtype_sitk:  # type: ignore[attr-defined]
                        image = sitk.Cast(image, merge_dtype_sitk)  # type: ignore[no-untyped-call]

                    # make sure we have numpy array
                    if isinstance(image, sitk.Image):
                        image = sitk.GetArrayFromImage(image)

                    # apply crop mask
                    if self.crop_mask is not None:
                        image = self.crop_mask * image
                    elif self.crop_bbox is not None:
                        x, y, width, height = self.crop_bbox
                        image = image[y : y + height, x : x + width]

                    # write OME-XML to the ImageDescription tag of the first page
                    if not ome_set:
                        description = self.omexml
                        ome_set = True

                    # write channel data
                    msg = (
                        f"Writing image for reader={reader_names[reader_index]} ({reader_index})"
                        f" channel={channel_name} ({channel_index})"
                    )
                    past_msg = msg.replace("Writing", "Wrote")
                    logger.trace(f"{msg} - {image.shape}...")
                    with MeasureTimer() as write_timer:
                        tif.write(image, subifds=self.subifds, description=description, **options)
                        logger.trace(f"{past_msg} pyramid index 0 in {write_timer()}")
                        if write_pyramid:
                            for pyramid_index in range(1, self.n_pyr_levels):
                                resize_shape = (self.pyr_levels[pyramid_index][0], self.pyr_levels[pyramid_index][1])
                                image = cv2.resize(image, resize_shape, cv2.INTER_LINEAR)
                                logger.trace(f"{msg} pyramid index {pyramid_index} - {image.shape}...")
                                tif.write(image, **options, subfiletype=1)
                                logger.trace(
                                    f"{past_msg} pyramid index {pyramid_index} in {write_timer(since_last=True)}"
                                )
        logger.trace(f"Exported OME-TIFF in {main_timer()}")
        # rename temporary file to final file
        retry(lambda: tmp_output_file_name.rename(output_file_name), PermissionError)()  # type: ignore[arg-type]
        logger.trace(f"Renamed tmp file to output file ({output_file_name})")
        return Path(output_file_name)

    def write(
        self,
        name: str,
        reader_names: list[str],
        output_dir: Path,
        tile_size: int = 512,
        as_uint8: bool = False,
        channel_ids: list[list[int] | None] | None = None,
    ) -> Path:
        """Write image."""
        return self.merge_write_image_by_plane(
            name, reader_names, output_dir, tile_size=tile_size, as_uint8=as_uint8, channel_ids=channel_ids
        )



"""Various image readers."""

from importlib.metadata import PackageNotFoundError, version

try:
    __version__ = version("image2image-io")
except PackageNotFoundError:
    __version__ = "uninstalled"

__author__ = "Lukasz G. Migas"
__email__ = "lukas.migas@yahoo.com"



"""Test reader."""

import numpy as np
import pytest

from image2image_io.readers import ArrayImageReader, get_simple_reader
from image2image_io.utils._test import get_test_files


@pytest.mark.parametrize("path", get_test_files("*.tiff"))
def test_get_simple_reader_tiff(path):
    reader = get_simple_reader(path)
    assert not reader.is_rgb, "Reader should not be rgb"
    assert reader.n_channels == 1, "Reader should have 1 channel"
    assert reader.reader_type == "image", "Reader should be image"


@pytest.mark.parametrize("path", get_test_files("*.png"))
def test_get_simple_reader_png(path):
    reader = get_simple_reader(path)
    assert reader.is_rgb, "Reader should not be rgb"
    assert reader.n_channels == 4, "Reader should have 1 channel"
    assert reader.reader_type == "image", "Reader should be image"


@pytest.mark.parametrize("path", get_test_files("*.jpg"))
def test_get_simple_reader_jpg(path):
    reader = get_simple_reader(path)
    assert reader.is_rgb, "Reader should not be rgb"
    assert reader.n_channels == 4, "Reader should have 1 channel"
    assert reader.reader_type == "image", "Reader should be image"


@pytest.mark.parametrize("path", get_test_files("*.imzML") + get_test_files("*.ibd"))
def test_get_simple_reader_imzml(path):
    reader = get_simple_reader(path)
    assert not reader.is_rgb, "Reader should not be rgb"
    assert reader.allow_extraction, "Reader should allow extraction"


@pytest.mark.parametrize("path", get_test_files("*.geojson"))
def test_get_simple_reader_geojson(path):
    reader = get_simple_reader(path)
    assert reader.reader_type == "shapes", "Reader should be shapes"


def test_crop_bbox_rgb(tmp_path):
    array = np.random.randint(0, 255, (1024, 1024, 3), dtype=np.uint8)
    reader = ArrayImageReader(tmp_path, array)
    assert reader.is_rgb, "Array should be rgb"

    cropped, _ = reader.crop_bbox(0, 512, 0, 512)
    assert cropped.shape == (512, 512, 3), "Cropped shape should be (512, 512, 3)"

    for array, _ in reader.crop_bbox_iter(0, 512, 0, 512):
        if array is None:
            continue
        assert array.shape == (512, 512), "Cropped shape should be (512, 512)"


@pytest.mark.parametrize("dtype", [np.float32, np.float64])
def test_crop_bbox_multichannel(tmp_path, dtype):
    array = np.random.random((7, 1024, 1024)).astype(dtype)
    reader = ArrayImageReader(tmp_path, array)
    assert not reader.is_rgb, "Array should not be rgb"

    cropped, _ = reader.crop_bbox(0, 512, 0, 512)
    assert cropped.shape == (7, 512, 512), "Cropped shape should be (512, 512, 3)"

    for array, _ in reader.crop_bbox_iter(0, 512, 0, 512):
        if array is None:
            continue
        assert array.shape == (512, 512), "Cropped shape should be (512, 512)"


def test_crop_polygon_rgb(tmp_path):
    array = np.random.randint(0, 255, (1024, 1024, 3), dtype=np.uint8)
    reader = ArrayImageReader(tmp_path, array)
    assert reader.is_rgb, "Array should be rgb"

    cropped, bbox = reader.crop_polygon([(0, 0), (512, 0), (512, 512), (0, 512)])
    assert cropped.shape == (512, 512, 3), "Cropped shape should be (512, 512, 3)"

    for array, _bbox in reader.crop_polygon_iter([(0, 0), (512, 0), (512, 512), (0, 512)]):
        if array is None:
            continue
        assert array.shape == (512, 512), "Cropped shape should be (512, 512)"


@pytest.mark.parametrize("dtype", [np.float32, np.float64])
def test_crop_polygon_multichannel(tmp_path, dtype):
    array = np.random.random((7, 1024, 1024)).astype(dtype)
    reader = ArrayImageReader(tmp_path, array)
    assert not reader.is_rgb, "Array should not be rgb"

    cropped, bbox = reader.crop_polygon([(0, 0), (512, 0), (512, 512), (0, 512)])
    assert cropped.shape == (7, 512, 512), "Cropped shape should be (512, 512, 3)"

    for array, _bbox in reader.crop_polygon_iter([(0, 0), (512, 0), (512, 512), (0, 512)]):
        if array is None:
            continue
        assert array.shape == (512, 512), "Cropped shape should be (512, 512)"



"""Lazy HDF5 wrapper around image data."""

from __future__ import annotations

import typing as ty
from contextlib import contextmanager
from pathlib import Path

import h5py
import numpy as np

from image2image_io.utils.utilities import reshape, reshape_batch


class LazyImageWrapper:
    """Lazy wrapper around hdf5 file."""

    array: np.ndarray | None = None

    def __init__(
        self, path: Path, key: str, channel_names: list[str], x_coordinates: np.ndarray, y_coordinates: np.ndarray
    ):
        self.path = path
        self.key = key
        self.channel_names = channel_names
        self.x_coordinates = x_coordinates
        self.y_coordinates = y_coordinates

    @contextmanager
    def open(self):
        """Open the file."""
        with h5py.File(self.path, "r") as fh:
            yield fh

    @contextmanager
    def lazy_array(self) -> ty.Generator[np.ndarray, None, None]:
        """Get reference to the peak's data without actually loading it into memory."""
        with self.open() as h5:
            yield h5[self.key]

    def __getitem__(self, item: int) -> np.ndarray:
        """Return item."""
        with self.lazy_array() as array:
            return reshape(self.x_coordinates, self.y_coordinates, array[:, item])

    @property
    def shape(self) -> tuple[int, int]:
        """Return shape of the array."""
        with self.lazy_array() as array:
            return array.shape  # type: ignore[return-value]

    @property
    def dtype(self) -> np.dtype:
        """Return dtype of the array."""
        with self.lazy_array() as peaks:
            return peaks.dtype  # type: ignore[no-any-return]

    def get_image(self) -> np.ndarray:
        """Return 3d version of the data."""
        if self.array is None:
            with self.open() as h5:
                array = np.zeros(self.shape, dtype=self.dtype)
                for sl in h5[self.key].iter_chunks():
                    array[sl] = h5[self.key][sl]
            self.array = reshape_batch(self.x_coordinates, self.y_coordinates, array)
        return self.array



"""Transform image."""

from __future__ import annotations

import typing as ty

import click
from koyo.click import Parameter, arg_split_int, arg_split_str, cli_parse_paths_sort, print_parameters, warning_msg
from loguru import logger

from image2image_io.enums import MaskOutputFmt


@click.group()
def transform():
    """Transform command."""


@click.option(
    "-W",
    "--overwrite",
    help="Overwrite existing data.",
    is_flag=True,
    default=None,
    show_default=True,
)
@click.option("-s", "--scene_index", help="Scene index to transform.", type=click.INT, default=None, show_default=True)
@click.option(
    "-f",
    "--fmt",
    help="Output format of the mask.",
    type=click.Choice(ty.get_args(MaskOutputFmt), case_sensitive=False),
    multiple=True,
    default=("binary",),
)
@click.option(
    "-o",
    "--output_dir",
    help="Path to directory where mask files will be saved.",
    default=".",  # cwd
    type=click.Path(exists=False, resolve_path=True, file_okay=False, dir_okay=True),
    show_default=True,
    required=True,
)
@click.option(
    "-T",
    "--transform",
    "transform_",
    help="Path to the i2r.json transformation file.",
    type=click.Path(exists=False, resolve_path=False, file_okay=True, dir_okay=False),
    show_default=True,
    required=True,
)
@click.option(
    "-m",
    "--mask",
    "mask_",
    help="Path to the mask file(s) that should be transformed. Can be GeoJSON or text format.",
    type=click.Path(exists=False, resolve_path=False, file_okay=True, dir_okay=False),
    show_default=True,
    required=True,
    multiple=True,
    callback=cli_parse_paths_sort,
)
@click.option(
    "-i",
    "--image",
    "image_",
    help="Path to the image file that the mask is transformed to. This could be OME-TIFF or IMS file (e.g. .d)",
    type=click.Path(exists=False, resolve_path=False, file_okay=True, dir_okay=True),
    show_default=True,
    required=True,
)
@transform.command(name="mask")
def mask(
    image_: str,
    mask_: list[str],
    output_dir: str,
    transform_: str,
    fmt: list[str],
    scene_index: int | None,
    overwrite: bool,
) -> None:
    """Transform GeoJSON or point mask using image2image transformation matrix (i2r.json)."""
    from image2image_io.utils.mask import transform_masks

    if not any(ext in transform_ for ext in (".i2r.json", ".i2r.toml")):
        raise ValueError("Only i2r.json or i2r.toml files are supported for mask transformation.")

    transform_masks(image_, mask_, output_dir, fmt, transform_, scene_index=scene_index, overwrite=overwrite)


@click.option(
    "-W",
    "--overwrite",
    help="Overwrite existing data.",
    is_flag=True,
    default=None,
    show_default=True,
)
@click.option(
    "-o",
    "--output_dir",
    help="Path to directory where mask files will be saved.",
    default=".",  # cwd
    type=click.Path(exists=False, resolve_path=True, file_okay=False, dir_okay=True),
    show_default=True,
    required=True,
)
@click.option(
    "-T",
    "--transform",
    "transform_",
    help="Path to the i2r.json transformation file.",
    type=click.Path(exists=False, resolve_path=False, file_okay=True, dir_okay=False),
    show_default=True,
    required=True,
)
@click.option(
    "-f",
    "--files",
    help="Path to the mask file(s) that should be transformed. Can be GeoJSON or text format.",
    type=click.Path(exists=False, resolve_path=False, file_okay=True, dir_okay=False),
    show_default=True,
    required=True,
    multiple=True,
    callback=cli_parse_paths_sort,
)
@transform.command(name="attachment")
def attachment(
    files: list[str],
    output_dir: str,
    transform_: str,
    overwrite: bool,
) -> None:
    """Transform GeoJSON or point mask using image2image transformation matrix (i2r.json)."""
    from image2image_io.utils.mask import transform_shapes_or_points

    if not any(ext in transform_ for ext in (".i2r.json", ".i2r.toml")):
        raise ValueError("Only i2r.json or i2r.toml files are supported for mask transformation.")

    transform_shapes_or_points(files, output_dir, transform_, overwrite=overwrite)


@click.option(
    "-W",
    "--overwrite",
    help="Overwrite existing data.",
    is_flag=True,
    default=None,
    show_default=True,
)
@click.option(
    "-R",
    "--resolution",
    help="Image resolution - override in case it's not specified in the file.",
    type=click.FLOAT,
    default=None,
    show_default=True,
)
@click.option(
    "--inverse",
    is_flag=True,
    help="Transform the image using the inverse transformation matrix (from moving to fixed).",
    show_default=True,
    required=False,
    default=False,
)
@click.option(
    "-s",
    "--suffix",
    type=click.STRING,
    help="Specify the suffix that goes at the end of the filename.",
    default="_transformed",
    show_default=True,
    required=True,
)
@click.option(
    "-N",
    "--channel_names",
    type=click.STRING,
    default=None,
    help="Specify a list of channel names to be used by the task. Names should be separated by commas. "
    "e.g. 'DAPI,CD3,CD20'",
    callback=arg_split_str,
    show_default=True,
    required=False,
)
@click.option(
    "-I",
    "--channel_ids",
    type=click.STRING,
    default=None,
    help="Specify a list of channel indices to be used by the task. Channel IDs should be separated by comma."
    "e.g. '1,2,4'.",
    callback=arg_split_int,
    show_default=True,
    required=False,
)
@click.option(
    "-u",
    "--as_uint8",
    is_flag=True,
    help="Convert to uint8. If not specified, the original data type will be used.",
    show_default=True,
    required=False,
)
@click.option(
    "-t",
    "--tile_size",
    help="Tile size.",
    type=click.Choice(["256", "512", "1024", "2048"], case_sensitive=False),
    default="512",
    show_default=True,
    required=False,
)
@click.option(
    "-o",
    "--output_dir",
    help="Path to directory where OME-TIFF files will be saved.",
    default=".",  # cwd
    type=click.Path(exists=False, resolve_path=True, file_okay=False, dir_okay=True),
    show_default=True,
    required=True,
)
@click.option(
    "-T",
    "--transform",
    "transform_",
    help="Path to the i2r.json transformation file.",
    type=click.Path(exists=False, resolve_path=False, file_okay=True, dir_okay=False),
    show_default=True,
    required=True,
)
@click.option(
    "-i",
    "--image",
    "image_",
    help="Path to the OME-TIFF file that should be converted.",
    type=click.Path(exists=False, resolve_path=False, file_okay=True, dir_okay=False),
    show_default=True,
    required=True,
    multiple=True,
    callback=cli_parse_paths_sort,
)
@transform.command(name="image")
def image(
    image_: list[str],
    output_dir: str,
    transform_: str,
    tile_size: int,
    as_uint8: bool,
    channel_ids: list[int] | None,
    channel_names: list[str | None],
    suffix: str = "_transformed",
    inverse: bool = True,
    resolution: float | None = None,
    overwrite: bool = False,
) -> None:
    """Transform OME-TIFF using image2image transformation matrix (i2r.json) or elastix (i2reg) transformation."""
    from image2image_io.writers import image_to_ome_tiff

    metadata = None
    if channel_ids and channel_names:
        if len(channel_ids) != len(channel_names):
            raise ValueError("Number of channel IDs and channel names must be equal.")
        metadata = {0: {"channel_ids": channel_ids, "channel_names": channel_names}}
    if len(image_) > 1 and metadata:
        warning_msg(
            "Multiple image files were specified wih a single set of metadata - it will be repeated for each image."
        )

    if any(ext in transform_ for ext in (".i2r.json", ".i2r.toml")):
        from image2image_io.utils.warp import ImageWarper

        transform_seq = ImageWarper(transform_, inv=inverse)
    else:
        try:
            from image2image_reg.elastix.transform_sequence import TransformSequence

            transform_seq = TransformSequence.from_path(transform_)
        except ImportError:
            raise ImportError("Please install image2image-reg to use i2reg transformations.")

    print_parameters(
        Parameter("Name", "-i/--image", image_),
        Parameter("Output directory", "-o/--output_dir", output_dir),
        Parameter("Channel ids", "-C/--channel_ids", channel_ids),
        Parameter("Write images as uint8", "--as_uint8/--no_as_uint8", as_uint8),
        Parameter("Overwrite", "-W/--overwrite", overwrite),
    )

    for image_path in image_:
        for key, scene_index, total, _, _ in image_to_ome_tiff(
            image_path,
            output_dir,
            as_uint8=as_uint8,
            tile_size=int(tile_size),
            metadata=metadata,
            transformer=transform_seq,
            overwrite=overwrite,
            suffix=suffix,
            resolution=resolution,
        ):
            logger.info(f"Transformed {key} {scene_index}/{total}")



"""Base image wrapper."""

from __future__ import annotations

import math
import typing as ty
import warnings
from pathlib import Path

import numpy as np
import zarr.storage
from koyo.typing import PathLike
from loguru import logger

from image2image_io.config import CONFIG
from image2image_io.enums import DEFAULT_TRANSFORM_NAME, DisplayType
from image2image_io.models.preprocess import NoopProcessor
from image2image_io.models.transform import TransformData
from image2image_io.utils.utilities import guess_rgb

if ty.TYPE_CHECKING:
    from image2image_io.writers.tiff_writer import Transformer

logger = logger.bind(src="Reader")


def check_if_open(path: Path) -> None:
    """Check if file is open."""
    import psutil

    # using psutil to check if file is open in the current process
    current_process_id = psutil.Process().pid
    proc = psutil.Process(current_process_id)
    for file in proc.open_files():
        if file.path == str(path):
            logger.warning(f"File {path} is still open in the current process...")


class BaseReader:
    """Base class for some of the other image readers."""

    _closed = False
    _pyramid = None
    _is_rgb: bool | None = None
    _array_dtype: np.dtype | None = None
    _array_shape: tuple[int, ...] | None = None
    _image_shape: tuple[int, int] | None = None
    _zstore: zarr.storage.TempStore | None = None
    _scene_ids: list[int] | None = None
    _resolution: float = 1.0
    _channel_names: list[str] | None = None
    _channel_colors: list[str] | None = None
    _channel_ids: list[int] | None = None

    auto_pyramid: bool | None = None
    reader: str = "base"
    reader_type: str = "image"
    display_type: DisplayType | None = None
    n_scenes: int = 1
    lazy: bool = False
    fh: ty.Any | None = None
    allow_extraction: bool = False

    def __init__(
        self, path: PathLike, key: str | None = None, reader_kws: dict | None = None, auto_pyramid: bool | None = None
    ):
        self.auto_pyramid = auto_pyramid
        # This is the direct path to the image
        self.path = Path(path)
        # This is the attribute we will use to identify the image
        self.key = key or self.path.name
        self.reader_kws = reader_kws or {}
        self.transform_data: TransformData = TransformData()
        self.transform_name = DEFAULT_TRANSFORM_NAME
        self.preprocessor = NoopProcessor()

    def __repr__(self) -> str:
        return (
            f"{self.__class__.__name__}<{self.key!r}; {self.path!r}; RGB={self.is_rgb}; {self.shape!r}"
            f" @ {self.resolution:.4f}>"
        )

    def find_channel_name(self, name: str) -> int | None:
        """Find cycle by name or part of the name."""
        if name in self.channel_names:
            return self.channel_names.index(name)
        for i, n in enumerate(self.channel_names):
            if name in n:
                return i
        return None

    def print_info(self) -> None:
        """Print information about the image."""
        print(f"Image: {self.path}")
        print(f"  - Resolution: {self.resolution}")
        print(f"  - Shape: {self.shape}")
        print(f"  - Channels: {self.n_channels}")
        print(f"  - RGB: {self.is_rgb}")
        for i, channel in enumerate(self.channel_names):
            print(f"  - Channel {i}: {channel}")
        print(f"  - Transform: {self.transform}")

    @property
    def scene_index(self) -> int:
        """Return scene index."""
        if self.reader_kws and "scene_index" in self.reader_kws:
            return self.reader_kws["scene_index"]
        return 0

    @property
    def transform(self) -> np.ndarray:
        """Return transform."""
        transform: np.ndarray = self.transform_data.transform.params
        return transform

    @transform.setter
    def transform(self, value: np.ndarray) -> None:
        assert value.shape == (3, 3)
        self.transform_data._transform = value

    def is_identity_transform(self) -> bool:
        """Return whether transform is identity."""
        if self.transform_data.transform:
            return np.allclose(self.transform_data.transform.params, np.eye(3))
        return np.allclose(self.transform, np.eye(3))

    @property
    def inv_resolution(self) -> float:
        """Return inverse resolution."""
        return 1 / self.resolution

    @property
    def shape(self) -> tuple[int, ...]:
        """Return shape of the image, including channels, etc."""
        if self._array_shape is None:
            self._array_shape = self.pyramid[0].shape  # TODO: this is wrong!
        return self._array_shape

    @shape.setter
    def shape(self, value: tuple | np.ndarray) -> None:
        """Set shape."""
        if self._array_shape is not None:
            warnings.warn("Shape is already set, overwriting it.", stacklevel=2)
        self._array_shape = tuple(value)

    @property
    def image_shape(self) -> tuple[int, int]:
        """Image shape."""
        if self._image_shape is None:
            from image2image_io.utils.utilities import get_shape_of_image

            self._image_shape = get_shape_of_image(self.shape)[-1]
        return self._image_shape

    @property
    def channel_names(self) -> list[str]:
        """Return channel names."""
        return self._channel_names

    def get_channel_names(self, split_rgb: bool | None = None) -> list[str]:
        """Get channel names."""
        split_rgb = split_rgb if split_rgb is not None else CONFIG.split_rgb
        if self.is_rgb:
            return ["R", "B", "B"] if split_rgb else ["RGB"]
        return self.channel_names

    @property
    def channel_ids(self) -> list[int]:
        """Return channel indices."""
        if self._channel_ids is None:
            channel_names = self._channel_names or self.channel_names
            self._channel_ids = list(range(len(channel_names)))
        return self._channel_ids

    @channel_ids.setter
    def channel_ids(self, value: list[int]) -> None:
        """Set channel indices."""
        if not isinstance(value, list):
            value = list(value)
        assert len(value) <= len(self._channel_names), (
            f"Too many channels selected: {len(value)} > {len(self._channel_names)}"
        )
        assert max(value) < len(self._channel_names), (
            f"Channel index out of range: {max(value)} > {len(self._channel_names)}"
        )
        self._channel_ids = value

    @property
    def scene_ids(self) -> list[int]:
        """Return channel indices."""
        if self._scene_ids is None:
            self._scene_ids = list(range(self.n_scenes))
        return self._scene_ids

    @scene_ids.setter
    def scene_ids(self, value: list[int]) -> None:
        """Set channel indices."""
        if not isinstance(value, list):
            value = list(value)
        assert len(value) <= self.n_scenes, f"Too many channels selected: {len(value)} > {self.n_scenes}"
        assert max(value) < self.n_scenes, f"Channel index out of range: {max(value)} > {self.n_scenes}"
        self._scene_ids = value

    @property
    def channel_colors(self) -> list[str] | None:
        """Return channel names."""
        return self._channel_colors

    def channel_to_index(self, channel: str) -> int:
        """Return index of a channel."""
        if channel in self.channel_names:
            return self.channel_names.index(channel)
        return -1

    @property
    def n_channels(self) -> int:
        """Return number of channels."""
        return len(self.channel_names)

    @property
    def dtype(self) -> np.dtype:
        """Return dtype."""
        if self._array_dtype is None:
            self._array_dtype = self.pyramid[0].dtype
        return self._array_dtype

    @property
    def is_rgb(self) -> bool:
        """Return whether image is RGB."""
        if self._is_rgb is None:
            self._is_rgb = guess_rgb(self.shape)
        return self._is_rgb

    @property
    def scale(self) -> tuple[float, float]:
        """Return scale."""
        return self.resolution, self.resolution

    @property
    def resolution(self) -> float:
        """Return resolution."""
        return self._resolution or 1.0

    @resolution.setter
    def resolution(self, value: float) -> None:
        self._resolution = value
        if self.transform_data:
            self.transform_data.moving_resolution = value

    @property
    def name(self) -> str:
        """Return name of the input path."""
        return self.path.name

    @property
    def display_name(self) -> str:
        """Retrieve display name from the path."""
        return self.path.stem

    @property
    def clean_key(self) -> str:
        """Key without extras."""
        return self.key[0:-4]

    @property
    def clean_name(self) -> str:
        """Return name of the input path."""
        name = self.name
        for ext in [
            ".ome",
            ".tiff",
            ".tif",
            ".png",
            ".jpg",
            ".svs",
            ".czi",
            ".d",
            ".imzML",
            ".tsf",
            ".tdf",
            ".qptiff",
            ".raw",
            ".intermediate",
        ]:
            name = name.replace(ext, "")
        return name

    def flat_array(
        self, channel_indices: list[int] | None = None, index: int = 0
    ) -> tuple[np.ndarray, tuple[int, int]]:
        """Return a flat array."""
        from image2image_io.utils.utilities import get_shape_of_image

        array = self.pyramid[index]
        if hasattr(array, "compute"):
            array = array.compute()

        if channel_indices is not None:
            channel_indices = list(channel_indices)
            assert all(channel_index in self.channel_ids for channel_index in channel_indices), "Invalid channel index"
        else:
            channel_indices = self.channel_ids

        # sub-select channels
        if array.ndim == 3:
            if self.is_rgb:
                array = array[:, :, channel_indices]
            else:
                array = array[channel_indices, :, :]

        # reshape array
        n_channels, _, shape = get_shape_of_image(array)
        if array.ndim == 3:
            array = array.reshape(-1, n_channels)
        else:
            array = array.reshape(-1, 1)
        return array, shape

    def close(self) -> None:
        """Close the file handle."""
        if self.fh and hasattr(self.fh, "close") and callable(self.fh.close):
            self.fh.close()
        if self.fh and hasattr(self.fh, "filehandle"):
            self.fh.filehandle.close()
        if self._zstore is not None and hasattr(self._zstore, "path") and self._zstore.path:
            zarr_path = Path(self._zstore.path)
            if zarr_path.exists():
                zarr.storage.atexit_rmtree(zarr_path)
                logger.trace(f"Removed temporary zarr store: {zarr_path}")
        self.fh = None
        self._pyramid = None
        # check_if_open(self.path)
        if not self._closed:
            logger.trace(f"Closed file handle '{self.path}'")
            self._closed = True

    @property
    def pyramid(self) -> list:
        """Pyramid."""
        if self._pyramid is None:
            self._pyramid = self.get_dask_pyr()
        return self._pyramid

    def get_dask_pyr(self) -> list[ty.Any]:
        """Get dask representation of the pyramid."""
        raise NotImplementedError("Must implement method")

    def get_thumbnail(self) -> tuple[np.ndarray, tuple[float, float]]:
        """Return thumbnail."""
        return self.pyramid[-1], self.scale_for_pyramid(-1)

    def crop_region(
        self, bbox_or_yx: np.ndarray | tuple[int, int, int, int]
    ) -> tuple[np.ndarray, tuple[int, int, int, int]]:
        """Crop region."""
        if isinstance(bbox_or_yx, tuple) and len(bbox_or_yx) == 4:
            return self.crop_bbox(*bbox_or_yx)
        return self.crop_polygon(bbox_or_yx)

    def crop_region_iter(
        self, bbox_or_yx: np.ndarray | tuple[int, int, int, int]
    ) -> ty.Generator[tuple[np.ndarray | None, tuple[int, int, int, int]], None, None]:
        """Crop region."""
        if isinstance(bbox_or_yx, tuple) and len(bbox_or_yx) == 4:
            yield from self.crop_bbox_iter(*bbox_or_yx)
        else:
            yield from self.crop_polygon_iter(bbox_or_yx)

    def crop_bbox(
        self,
        left: int,
        right: int,
        top: int,
        bottom: int,
        array: np.ndarray | None = None,
        multiply: bool = True,
        apply: bool = True,
    ) -> tuple[np.ndarray, tuple[int, int, int, int]]:
        """Crop image."""
        inv_resolution = self.inv_resolution if multiply else 1
        top, bottom = sorted([top, bottom])
        left, right = sorted([left, right])
        left = math.floor(left * inv_resolution)
        right = math.ceil(right * inv_resolution)
        top = math.floor(top * inv_resolution)
        bottom = math.ceil(bottom * inv_resolution)

        # get array
        if array is None:
            array = self.pyramid[0]
        if array.ndim == 2:
            array_ = array[top:bottom, left:right]
        elif array.ndim == 3:
            channel_axis, _ = self.get_channel_axis_and_n_channels()
            if channel_axis == 0:
                array_ = array[:, top:bottom, left:right]
            elif channel_axis == 1:
                array_ = array[top:bottom, :, left:right]
            elif channel_axis == 2:
                array_ = array[top:bottom, left:right, :]
            else:
                raise ValueError(f"Array has unsupported shape: {array.shape}")
        else:
            raise ValueError(f"Array has unsupported shape: {array.shape}")
        # check whether an array is dask array - if so, we need to compute it
        if hasattr(array_, "compute") and apply:
            array_ = array_.compute()
        return array_, (left, right, top, bottom)  # type: ignore[no-any-return]

    def crop_bbox_iter(
        self,
        left: int,
        right: int,
        top: int,
        bottom: int,
        multiply: bool = True,
        apply: bool = True,
    ) -> ty.Generator[tuple[np.ndarray | None, tuple[int, int, int, int]], None, None]:
        """Crop image."""
        inv_resolution = self.inv_resolution if multiply else 1
        top, bottom = sorted([top, bottom])
        left, right = sorted([left, right])
        left = math.floor(left * inv_resolution)
        right = math.ceil(right * inv_resolution)
        top = math.floor(top * inv_resolution)
        bottom = math.ceil(bottom * inv_resolution)
        yield None, (left, right, top, bottom)

        for channel_id in range(self.n_channels):
            array = self.get_channel(channel_id, split_rgb=True)
            array_ = array[top:bottom, left:right]
            # check whether an array is dask array - if so, we need to compute it
            if hasattr(array_, "compute") and apply:
                array_ = array_.compute()
            yield array_, (left, right, top, bottom)  # type: ignore[no-any-return]

    def crop_polygon(self, yx: np.ndarray) -> tuple[np.ndarray, tuple[int, int, int, int]]:
        """Crop image using polygon, however, try to minimize memory usage."""
        import cv2

        # reverse yx to xy
        yx = np.asarray(yx)
        xy = yx[:, ::-1] * self.inv_resolution
        xy = np.round(xy).astype(np.int32)

        # get left/right/top/bottom from polygon
        h, w = self.image_shape
        left, bottom = np.min(xy, axis=0)
        left = np.max([0, left])
        bottom = np.max([0, bottom])
        right, top = np.max(xy, axis=0)
        right = np.min([w, right])
        top = np.min([h, top])
        top, bottom = sorted([top, bottom])
        left, right = sorted([left, right])

        cropped_array, _ = self.crop_bbox(left, right, top, bottom, multiply=False, apply=True)

        # Adjust polygon coordinates for the cropped region
        xy_cropped = xy - np.array([left, top])

        # Create mask only for the cropped region
        cropped_shape = self.get_image_shape_for_shape(cropped_array.shape)
        mask = np.zeros(cropped_shape, dtype=np.uint8)
        cv2.fillPoly(mask, pts=[xy_cropped], color=np.iinfo(np.uint8).max)
        mask = mask.astype(bool)

        # Apply mask
        if cropped_array.ndim == 2:
            cropped_array[~mask] = 0
        else:
            channel_axis, _ = self.get_channel_axis_and_n_channels()
            if channel_axis == 0:
                cropped_array *= mask
            elif channel_axis in (1, 2):  # Assuming CxHxW, HxWxC, or HxW
                cropped_array *= mask[:, :, None]
            else:
                raise ValueError(f"Unsupported array shape: {cropped_array.shape}")

        # If the array is a dask array, compute it
        if hasattr(cropped_array, "compute"):
            cropped_array = cropped_array.compute()
        return cropped_array, (left, right, top, bottom)

    def crop_polygon_iter(
        self, yx: np.ndarray
    ) -> ty.Generator[tuple[np.ndarray | None, tuple[int, int, int, int]], None, None]:
        """Polygon iterator."""
        import cv2

        # reverse yx to xy
        yx = np.asarray(yx)
        xy = yx[:, ::-1] * self.inv_resolution
        xy = np.round(xy).astype(np.int32)

        # get left/right/top/bottom from polygon
        h, w = self.image_shape
        left, bottom = np.min(xy, axis=0)
        left = np.max([0, left])
        bottom = np.max([0, bottom])
        right, top = np.max(xy, axis=0)
        right = np.min([w, right])
        top = np.min([h, top])
        top, bottom = sorted([top, bottom])
        left, right = sorted([left, right])

        # Adjust polygon coordinates for the cropped region
        xy_cropped = xy - np.array([left, top])
        yield None, (left, right, top, bottom)

        mask = None
        for channel_id in range(self.n_channels):
            array = self.get_channel(channel_id, split_rgb=True)
            cropped_array, _ = self.crop_bbox(left, right, top, bottom, array=array, multiply=False, apply=True)

            # create mask only once
            if mask is None:
                # Create mask only for the cropped region
                cropped_shape = self.get_image_shape_for_shape(cropped_array.shape)
                mask = np.zeros(cropped_shape, dtype=np.uint8)
                cv2.fillPoly(mask, pts=[xy_cropped], color=np.iinfo(np.uint8).max)
                mask = mask.astype(bool)

            # apply mask
            cropped_array *= mask
            if hasattr(cropped_array, "compute"):
                cropped_array = cropped_array.compute()
            yield cropped_array, (left, right, top, bottom)

    def crop_polygon_mask(self, yx: np.ndarray) -> tuple[np.ndarray, tuple[int, int, int, int]]:
        """Crop image."""
        import cv2

        # reverse yx to xy
        xy = yx[:, ::-1] * self.inv_resolution
        xy = np.round(xy).astype(np.int32)
        # get left/right/top/bottom from polygon
        h, w = self.image_shape
        left, bottom = np.min(xy, axis=0)
        left = np.max([0, left])
        bottom = np.max([0, bottom])
        right, top = np.max(xy, axis=0)
        right = np.min([w, right])
        top = np.min([h, top])
        top, bottom = sorted([top, bottom])
        left, right = sorted([left, right])

        # get mask
        mask = np.zeros(self.image_shape, dtype=np.uint8)
        mask = cv2.fillPoly(mask, pts=[xy], color=np.iinfo(np.uint8).max)
        mask = mask.astype(bool)

        # get array
        array = self.pyramid[0]
        array_ = array.copy()
        if array.ndim == 2:
            array_[mask] = 0
        elif array_.ndim == 3:
            channel_axis, _ = self.get_channel_axis_and_n_channels()
            if channel_axis == 0:
                array_ = array_ * mask
            elif channel_axis in [1, 2]:
                array_ = array_ * mask[:, :, None]
            else:
                raise ValueError(f"Array has unsupported shape: {array_.shape}")
        # check whether an array is dask array - if so, we need to compute it
        if hasattr(array_, "compute"):
            array_ = array_.compute()
        array_, _ = self.crop_bbox(left, right, top, bottom, array=array_, multiply=False)
        return array_, (left, right, top, bottom)

    def crop_polygon_mask_iter(
        self, yx: np.ndarray
    ) -> ty.Generator[tuple[np.ndarray | None, tuple[int, int, int, int]], None, None]:
        """Crop image."""
        import cv2

        # reverse yx to xy
        xy = yx[:, ::-1] * self.inv_resolution
        xy = np.round(xy).astype(np.int32)
        # get left/right/top/bottom from polygon
        h, w = self.image_shape
        left, bottom = np.min(xy, axis=0)
        left = np.max([0, left])
        bottom = np.max([0, bottom])
        right, top = np.max(xy, axis=0)
        right = np.min([w, right])
        top = np.min([h, top])
        top, bottom = sorted([top, bottom])
        left, right = sorted([left, right])

        # get mask
        mask = np.zeros(self.image_shape, dtype=np.uint8)
        mask = cv2.fillPoly(mask, pts=[xy], color=np.iinfo(np.uint8).max)
        mask = mask.astype(bool)
        yield None, (left, right, top, bottom)

        # get array
        for channel_id in range(self.n_channels):
            array_ = self.get_channel(channel_id)
            array_[mask] = 0
            # check whether an array is dask array - if so, we need to compute it
            if hasattr(array_, "compute"):
                array_ = array_.compute()
            array_, _ = self.crop_bbox(left, right, top, bottom, array=array_, multiply=False)
            yield array_, (left, right, top, bottom)

    def warp(self, array: np.ndarray, affine: np.ndarray | None = None) -> np.ndarray:
        """Warp array.

        Parameters
        ----------
        array : np.ndarray
            Array to warp.
        """
        from image2image_io.utils.mask import transform_mask

        if affine is None:
            affine = self.transform_data.compute(yx=True, px=True).params
        transformed_mask = transform_mask(array, affine, self.image_shape)
        return transformed_mask

    def get_channel_axis_and_n_channels(self, shape: tuple[int, ...] | None = None) -> tuple[int | None, int]:
        """Return channel axis and number of channels."""
        if shape is None:
            shape = self.shape
        ndim = len(shape)
        # 2D images will be returned as they are
        if ndim == 2:
            channel_axis = None
            n_channels = 1
        # 3D images will be split into channels
        elif ndim == 3:
            if shape[2] in [3, 4]:  # rgb
                channel_axis = 2
                n_channels = shape[2]
            # elif np.argmin(shape) == 2:
            #     channel_axis = 2
            #     n_channels = shape[2]
            else:  # channels first
                channel_axis = 0
                n_channels = shape[0]
        else:
            raise ValueError(f"Array has unsupported shape: {shape}")
        return channel_axis, n_channels

    def get_image_shape_for_shape(self, shape: tuple | None = None) -> tuple[int, int]:
        """Return shape of an image for a given shape."""
        if shape is None:
            shape = self.shape
        channel_axis, n_channels = self.get_channel_axis_and_n_channels(shape)
        if channel_axis is None or (self.is_rgb and not CONFIG.split_rgb):
            return shape[:2]
        if channel_axis == 0:
            return shape[1:]
        elif channel_axis == 1:
            return shape[0], shape[2]
        elif channel_axis == 2:
            return shape[:2]
        raise ValueError(f"Array has unsupported shape: {shape}")

    def get_channel(self, index: int, pyramid: int = 0, split_rgb: bool | None = None) -> np.ndarray:
        """Return channel."""
        split_rgb = split_rgb if split_rgb is not None else CONFIG.split_rgb

        array: np.ndarray = self.pyramid[pyramid]
        channel_axis, n_channels = self.get_channel_axis_and_n_channels()
        # if channel_axis in [None, 2] or (self.is_rgb and (not CONFIG.split_rgb and not split_rgb)):
        #     return self.preprocessor("rgb", array)
        if channel_axis is None:
            return self.preprocessor("gray", array)
        if channel_axis == 0:
            return self.preprocessor(self.channel_names[index], array[index])
        elif channel_axis == 1:
            return self.preprocessor(self.channel_names[index], array[:, index])
        elif channel_axis == 2:
            if self.is_rgb:
                if CONFIG.split_rgb or split_rgb:
                    return self.preprocessor("RGB"[index], array[:, :, index])
                return self.preprocessor("RGB", array)
            return self.preprocessor(self.channel_names[index], array[:, :, index])
        raise ValueError(f"Array has unsupported shape: {array.shape}")

    def get_channel_pyramid(self, index: int) -> list[np.ndarray]:
        """Return channel pyramid."""
        array = self.pyramid
        channel_axis, n_channels = self.get_channel_axis_and_n_channels()
        if channel_axis is None or (self.is_rgb and not CONFIG.split_rgb):
            return array
        if channel_axis == 0:
            return [a[index] for a in array]
        elif channel_axis == 1:
            return [a[:, index] for a in array]
        elif channel_axis == 2:
            return [a[:, :, index] for a in array]
        else:
            raise ValueError("Could not retrieve channel pyramid.")

    @property
    def n_in_pyramid(self) -> int:
        """Return number of images in the pyramid."""
        return len(self.pyramid)

    @property
    def n_in_pyramid_quick(self) -> int | None:
        """Return number of images in the pyramid, if pyramid is available."""
        if self._pyramid is None:
            return None
        return self.n_in_pyramid

    def scale_for_pyramid(self, pyramid: int = 0) -> tuple[float, float]:
        """Return scale for pyramid."""
        if pyramid < 0:
            pyramid = list(range(self.n_in_pyramid))[pyramid]
            # pyramid = range(self.n_in_pyramid)[pyramid] + 1
        resolution = self.resolution * 2**pyramid
        return resolution, resolution

    def __del__(self):
        """Close the file handle."""
        self.close()

    def to_ome_tiff(
        self,
        path: PathLike,
        as_uint8: bool = False,
        tile_size: int = 512,
        channel_ids: list[int] | None = None,
        channel_names: list[str] | None = None,
        transformer: Transformer | None = None,
        overwrite: bool = False,
    ) -> Path:
        """Write image as OME-TIFF."""
        from image2image_io.writers import write_ome_tiff_alt

        return write_ome_tiff_alt(
            path,
            self,
            as_uint8=as_uint8,
            channel_names=channel_names,
            channel_ids=channel_ids,
            tile_size=tile_size,
            overwrite=overwrite,
            transformer=transformer,
        )

    def to_writer(self):
        """Get instance of writer."""
        from image2image_io.writers.tiff_writer import OmeTiffWriter

        writer = OmeTiffWriter(self, transformer=None)
        return writer


class DummyReader(BaseReader):
    """Dummy reader."""

    def __init__(self, resolution: float, channel_names: list[str], shape: tuple[int, ...], dtype: np.dtype):
        super().__init__("", "", auto_pyramid=False)
        self._channel_names = channel_names
        self._resolution = resolution
        self._array_dtype = dtype
        self.shape = shape

    def pyramid(self) -> list[np.ndarray]:
        """Pyramid."""
        return []



"""Warp utilities."""

from __future__ import annotations

from pathlib import Path

import dask.array
import numpy as np
import SimpleITK as sitk
from koyo.timer import MeasureTimer
from koyo.typing import PathLike
from loguru import logger
from tqdm import trange


def get_affine_from_config(
    config_or_path: PathLike | dict, yx: bool = True, px: bool = True, inv: bool = False
) -> tuple[np.ndarray, tuple[int, int], float]:
    """Get affine transformation matrix from config."""
    from koyo.json import read_json_data
    from koyo.toml import read_toml_data

    if not isinstance(config_or_path, dict):
        config_or_path = Path(config_or_path)
        if not config_or_path.exists():
            raise FileNotFoundError(f"Config file not found: {config_or_path}")
        if config_or_path.suffix == ".toml":
            config = read_toml_data(config_or_path)
        else:
            config = read_json_data(config_or_path)
    else:
        config = config_or_path

    # create matrix query
    key = "matrix"
    key += "_yx" if yx else "_xy"
    key += "_px" if px else "_um"
    key += "_inv" if inv else ""

    # get affine transformation
    affine_inv = config[key]
    affine_inv = np.asarray(affine_inv)
    # get output image shape
    fixed_paths = config["fixed_paths"]
    if not fixed_paths:
        raise ValueError("Fixed paths not found in config.")
    image_shapes = [tuple(t["image_shape"]) for t in fixed_paths]
    if len(set(image_shapes)) != 1:
        raise ValueError("Different image shapes found in config.")
    image_shape: tuple[int, int] = tuple(image_shapes[0])
    pixel_sizes_um = [float(t["pixel_size_um"]) for t in fixed_paths]
    if len(set(pixel_sizes_um)) != 1:
        raise ValueError("Different pixel sizes found in config.")
    pixel_size_um: float = pixel_sizes_um[0]
    return affine_inv, image_shape, pixel_size_um


def warp_path(config_path: PathLike, from_transform: PathLike) -> np.ndarray:
    """Warp image with image2image transformation matrix."""
    import cv2
    from scipy.ndimage import affine_transform

    from image2image_io.readers import get_simple_reader

    # load affine matrix
    affine_inv, output_shape, pixel_size = get_affine_from_config(config_path)

    if not Path(from_transform).exists():
        raise FileNotFoundError(f"File not found: {from_transform}")

    # load readers
    from_reader = get_simple_reader(from_transform)

    # due to a limitation in the opencv implementation, we need to use scipy if the image is too large
    use_cv2 = max(max(from_reader.image_shape), max(output_shape)) < 32767
    warped = []
    for channel in trange(from_reader.n_channels, desc="Warping images..."):
        img = from_reader.get_channel(channel)
        if use_cv2:
            if isinstance(img, dask.array.Array):
                img = img.T.compute()
            warp_img = cv2.warpAffine(img, np.linalg.inv(affine_inv)[:2, :], output_shape[::-1]).T
        else:
            warp_img = affine_transform(img, affine_inv, order=1, output_shape=output_shape)
        warped.append(warp_img)

    # stack image
    warped = np.dstack(warped)
    # ensure that RGB remains RGB but AF remain AF
    if warped.ndim == 3 and np.argmin(warped.shape) == 2 and not from_reader.is_rgb:
        warped = np.moveaxis(warped, 2, 0)
    return warped


def warp(affine_inv: np.ndarray, output_shape: tuple[int, int], image: np.ndarray, order: int = 1) -> np.ndarray:
    """Warp image."""
    import cv2
    from scipy.ndimage import affine_transform

    use_cv2 = max(max(image.shape), max(output_shape)) < 32767
    if use_cv2:
        logger.debug("Using cv2 for warping.")
        if isinstance(image, dask.array.Array):
            image = image.compute()
        warped = cv2.warpAffine(
            image.T,
            np.linalg.inv(affine_inv)[:2, :],
            output_shape,
            flags=cv2.INTER_NEAREST if order == 0 else cv2.INTER_LINEAR,
        ).T
    else:
        warped = affine_transform(image, affine_inv, order=order, output_shape=output_shape)
    return warped


class ImageWarper:
    """Image warper class."""

    def __init__(self, config_or_path: PathLike | dict, yx: bool = True, inv: bool = False):
        """Initialize."""
        self.config_or_path = Path(config_or_path) if not isinstance(config_or_path, dict) else config_or_path
        self.affine_inv, self.output_size_yx, pixel_size = get_affine_from_config(self.config_or_path, yx=yx, inv=inv)
        self.output_size = self.output_size_yx[::-1]  # x, y
        self.output_spacing = (pixel_size, pixel_size)

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}<shape(xy)={self.output_size}; pixel_size(xy)={self.output_spacing}>"

    def __call__(self, image: np.ndarray | sitk.Image) -> np.ndarray | sitk.Image:
        """Warp image."""
        with MeasureTimer() as timer:
            is_sitk = isinstance(image, sitk.Image)
            if is_sitk:
                image = sitk.GetArrayFromImage(image)
            image = warp(self.affine_inv, self.output_size_yx, image)
            assert image.shape == self.output_size_yx, f"Image shape mismatch: {image.shape} != {self.output_size_yx}"
            if is_sitk:
                image = sitk.GetImageFromArray(image)
            logger.trace(f"Warped image in {timer()}.")
        return image



"""Init."""

from __future__ import annotations

import typing as ty
from pathlib import Path

import numpy as np
from koyo.secret import hash_obj
from koyo.system import IS_MAC
from koyo.typing import PathLike
from loguru import logger

from image2image_io.config import CONFIG
from image2image_io.exceptions import UnsupportedFileFormatError
from image2image_io.models.transform import TransformData
from image2image_io.readers._base_reader import BaseReader
from image2image_io.readers.array_reader import ArrayImageReader
from image2image_io.readers.coordinate_reader import CoordinateImageReader, LazyCoordinateImageReader
from image2image_io.readers.czi_reader import CziImageReader, CziSceneImageReader
from image2image_io.readers.shapes_reader import ShapesReader, is_txt_and_has_columns
from image2image_io.readers.tiff_reader import TiffImageReader
from image2image_io.utils.utilities import get_yx_coordinates_from_shape, reshape, reshape_batch
from image2image_io.wrapper import ImageWrapper

__all__ = [
    "ArrayImageReader",
    "BaseReader",
    "CoordinateImageReader",
    "CziImageReader",
    "CziSceneImageReader",
    "LazyCoordinateImageReader",
    "ShapesReader",
    "TiffImageReader",
    "get_key",
    "is_supported",
    "read_data",
    "sanitize_path",
    "sanitize_read_path",
]


IMAGE_EXTENSIONS = [".jpg", ".jpeg", ".png"]
TIFF_EXTENSIONS = [
    ".scn",
    ".ome.tiff",
    ".tif",
    ".tiff",
    ".svs",
    ".ndpi",
    ".qptiff",
    ".qptiff.raw",
    ".qptiff.intermediate",
]
CZI_EXTENSIONS = [".czi"]
BRUKER_EXTENSIONS = [".tsf", ".tdf", ".d"]
IMZML_EXTENSIONS = [".imzml", ".ibd"]
H5_EXTENSIONS = [".h5", ".hdf5"]
IMSPY_EXTENSIONS = [".data"]
NPY_EXTENSIONS = [".npy"]
NPZ_EXTENSIONS = [".npz"]
GEOJSON_EXTENSIONS = [".geojson", ".json"]
POINTS_EXTENSIONS = [".csv", ".txt", ".parquet"]
SUPPORTED_IMAGE_FORMATS = [
    *IMAGE_EXTENSIONS,
    *TIFF_EXTENSIONS,
    *CZI_EXTENSIONS,
]


def sanitize_path(path: PathLike) -> Path:
    """Sanitize a path, so it has a unified format across models."""
    path = Path(path).resolve()
    if path.is_file():
        if path.suffix in [".tsf", ".tdf"]:
            path = path.parent
        elif path.name == "dataset.metadata.h5":
            path = path.parent
    return path


def get_alternative_path(path: PathLike) -> Path:
    """Retrieve alternative name."""
    path = Path(path)
    if path.is_dir() and path.suffix == ".data":
        return path / "dataset.metadata.h5"
    return path


def sanitize_read_path(path: PathLike, raise_on_error: bool = True) -> Path | None:
    """Sanitize the path that can be read by the reader."""
    path = Path(path)
    if not path.exists() and raise_on_error:
        raise FileNotFoundError(f"File does not exist: {path}")
    suffix = path.suffix.lower()
    if suffix not in (
        TIFF_EXTENSIONS
        + IMAGE_EXTENSIONS
        + CZI_EXTENSIONS
        + NPY_EXTENSIONS
        + NPZ_EXTENSIONS
        + BRUKER_EXTENSIONS
        + IMZML_EXTENSIONS
        + H5_EXTENSIONS
        + IMSPY_EXTENSIONS
        + GEOJSON_EXTENSIONS
        + POINTS_EXTENSIONS
        + [".tmp"]
    ):
        if raise_on_error:
            raise ValueError(f"Unsupported file format: {path.suffix} ({path})")
        return None
    if suffix in BRUKER_EXTENSIONS:
        if path.suffix == ".d":
            if (path / "analysis.tsf").exists():
                path = path / "analysis.tsf"
            else:
                path = path / "analysis.tdf"
    elif suffix in IMSPY_EXTENSIONS:
        path = path / "dataset.metadata.h5"
    return path


def read_data(
    path: PathLike,
    wrapper: ImageWrapper | None = None,
    is_fixed: bool = False,
    transform_data: TransformData | None = None,
    resolution: float | None = None,
    split_czi: bool | None = None,
    reader_kws: dict[str, ty.Any] | None = None,
) -> tuple[ImageWrapper, list[str], dict[Path, Path]]:
    """Read image data."""
    path = Path(path)
    path = sanitize_read_path(path)  # type: ignore[assignment]
    if not path:
        raise UnsupportedFileFormatError("Could not sanitize path - are you sure this file is supported?")

    readers: dict[str, BaseReader]
    name = path.name
    original_path = path
    scene_index = reader_kws.get("scene_index", None) if reader_kws else None
    path, readers = get_reader(path, split_czi=split_czi, scene_index=scene_index)

    # add transformation information if provided
    if transform_data is not None:
        for reader_ in readers.values():
            reader_.transform_data = transform_data
            reader_.transform_name = name
    # add resolution information if provided
    if resolution is not None:
        for reader_ in readers.values():
            reader_.resolution = resolution
    # specify whether the model is fixed
    if hasattr(readers, "is_fixed"):
        readers.is_fixed = is_fixed

    # initialize image wrapper, unless it's been provided
    if wrapper is None:
        wrapper = ImageWrapper(None)
    # add readers to the wrapper
    just_added = []
    for reader_ in readers.values():
        wrapper.add(reader_)
        just_added.append(reader_.key)
    return wrapper, just_added, {original_path: path}


def get_key(path: Path, scene_index: int | None = None) -> str:
    """Return representative key."""
    name = path.name
    hash_of_path = hash_obj(path, n_in_hash=3)
    if name in ["dataset.metadata.h5", "analysis.tsf", "analysis.tdf"]:
        name = path.parent.name
    if scene_index is not None:
        name = f"Scene={scene_index}; {name}"
    return f"{name}-{hash_of_path}"


def is_supported(path: PathLike, raise_on_error: bool = True) -> bool:
    """Check whether the file is supported."""
    path = Path(path)
    path = sanitize_read_path(path, raise_on_error)  # type: ignore[assignment]
    if not path:
        raise UnsupportedFileFormatError("Could not sanitize path - are you sure this file is supported?")
    return path is not None


def get_reader(
    path: Path,
    split_czi: bool | None = None,
    split_roi: bool | None = None,
    quick: bool = False,
    scene_index: int | None = None,
) -> tuple[Path, dict[str, BaseReader]]:
    """Get reader for the specified path."""
    path = Path(path)
    path = sanitize_read_path(path)  # type: ignore[assignment]
    if not path:
        raise UnsupportedFileFormatError("Could not sanitize path - are you sure this file is supported?")

    split_czi = split_czi if split_czi is not None else CONFIG.split_czi
    split_roi = split_roi if split_roi is not None else CONFIG.split_roi
    include_all = True

    if scene_index is not None:
        split_czi = split_roi = True
        include_all = False

    readers: dict[str, BaseReader]
    suffix = path.suffix.lower()
    if suffix in TIFF_EXTENSIONS:
        CONFIG.trace(f"Reading TIFF file: {path}")
        path, readers = _read_tiff(path)
    elif suffix in CZI_EXTENSIONS:
        if split_czi and _check_multi_scene_czi(path):
            logger.trace(f"Reading multi-scene CZI file: {path}")
            path, readers = _read_multi_scene_czi(path, scene_index=scene_index)
        else:
            CONFIG.trace(f"Reading single-scene CZI file: {path}")
            path, readers = _read_single_scene_czi(path)
    elif suffix in IMAGE_EXTENSIONS:
        CONFIG.trace(f"Reading image file: {path}")
        path, readers = _read_image(path)
    elif suffix in NPY_EXTENSIONS:
        CONFIG.trace(f"Reading NPY file: {path}")
        path, readers = _read_npy_coordinates(path)
    elif suffix in NPZ_EXTENSIONS:
        CONFIG.trace(f"Reading NPY file: {path}")
        path, readers = _read_npz_coordinates(path)
    elif suffix in BRUKER_EXTENSIONS:
        CONFIG.trace(f"Reading Bruker file: {path}")
        if IS_MAC or split_roi:
            path, readers = _read_tsf_tdf_coordinates(path, split_roi, scene_index=scene_index, include_all=include_all)
        else:
            path, readers = _read_tsf_tdf_reader(path)
    elif suffix in IMZML_EXTENSIONS:
        CONFIG.trace(f"Reading imzML file: {path}")
        path, readers = _read_imzml_reader(path)
    elif suffix in H5_EXTENSIONS + IMSPY_EXTENSIONS:
        CONFIG.trace(f"Reading HDF5 file: {path}")
        if path.suffix == ".data":
            path = path / "dataset.metadata.h5"
        if path.name.startswith("dataset.metadata"):
            path, readers = _read_metadata_h5_coordinates(path)
        elif _is_h5_mask(path):
            path, readers = _read_mask_h5(path)
        else:
            path, readers = _read_centroids_h5_coordinates_lazy(path)
    elif suffix in GEOJSON_EXTENSIONS + POINTS_EXTENSIONS:
        if (
            suffix in GEOJSON_EXTENSIONS
            or is_txt_and_has_columns(path, ["vertex_x", "vertex_y"], [("cell", "cell_id", "shape", "shape_name")])
            or is_txt_and_has_columns(
                path, ["x", "y"], [("cell", "cell_id", "shape", "shape_name")], (np.dtype("O"), np.dtype("S"))
            )
        ):
            CONFIG.trace(f"Reading shape file: {path}")
            path, readers = _read_shapes(path)
        else:
            CONFIG.trace(f"Reading points file: {path}")
            path, readers = _read_points(path)  # type: ignore
    else:
        raise UnsupportedFileFormatError(f"Unsupported file format: '{suffix}'")
    return path, readers


def get_simple_reader(
    path: PathLike,
    init_pyramid: bool = True,
    auto_pyramid: bool = True,
    quick: bool = False,
    quiet: bool = False,
    scene_index: int | None = None,
) -> BaseReader:
    """Get simple reader."""
    quiet_ = CONFIG.quiet
    CONFIG.quiet = quiet
    init_pyramid_ = CONFIG.init_pyramid
    CONFIG.init_pyramid = init_pyramid
    auto_pyramid_ = CONFIG.auto_pyramid
    CONFIG.auto_pyramid = auto_pyramid
    path, readers = get_reader(path, split_czi=False, quick=quick, scene_index=scene_index)
    CONFIG.init_pyramid = init_pyramid_
    CONFIG.auto_pyramid = auto_pyramid_
    CONFIG.quiet = quiet_
    return next(iter(readers.values()))


def _check_multi_scene_czi(path: PathLike) -> bool:
    """Check whether this is a multi-scene CZI file."""
    from image2image_io.readers._czi import CziSceneFile

    path = Path(path)
    return bool(CziSceneFile.get_num_scenes(path) > 1)


def _read_shapes(path: PathLike) -> tuple[Path, dict[str, ShapesReader]]:
    """Read GeoJSON file."""
    from image2image_io.readers.shapes_reader import ShapesReader

    path = Path(path)
    assert path.exists(), f"File does not exist: {path}"
    key = get_key(path)
    return path, {path.name: ShapesReader(path, key=key)}


def _read_points(path: PathLike) -> tuple[Path, dict[str, ShapesReader]]:
    """Read GeoJSON file."""
    from image2image_io.readers.points_reader import PointsReader

    path = Path(path)
    assert path.exists(), f"File does not exist: {path}"
    key = get_key(path)
    return path, {path.name: PointsReader(path, key=key)}


def _read_single_scene_czi(path: PathLike, **kwargs: ty.Any) -> tuple[Path, dict[str, CziImageReader]]:
    """Read CZI file."""
    from image2image_io.readers.czi_reader import CziImageReader

    path = Path(path)
    assert path.exists(), f"File does not exist: {path}"
    key = get_key(path)
    return path, {path.name: CziImageReader(path, key=key, **kwargs)}


def _read_multi_scene_czi(
    path: PathLike, scene_index: int | None = None
) -> tuple[Path, dict[str, CziSceneImageReader]]:
    """Read CZI file."""
    from image2image_io.readers._czi import CziSceneFile
    from image2image_io.readers.czi_reader import CziSceneImageReader

    path = Path(path)
    assert path.exists(), f"File does not exist: {path}"
    n = CziSceneFile.get_num_scenes(path)
    scenes = list(range(n))
    if scene_index is not None:
        if scene_index not in scenes:
            raise ValueError(f"Scene index {scene_index} not found in the file")
        scenes = [scene_index]

    CONFIG.trace(f"Found {n} scenes in CZI file: {path}")
    return path, {
        f"S{i}_{path.name}": CziSceneImageReader(path, scene_index=i, key=get_key(path, scene_index=i)) for i in scenes
    }


def _read_tiff(path: PathLike) -> tuple[Path, dict[str, TiffImageReader]]:
    """Read TIFF file."""
    from image2image_io.readers.tiff_reader import TiffImageReader

    path = Path(path)
    assert path.exists(), f"File does not exist: {path}"
    key = get_key(path)
    return path, {path.name: TiffImageReader(path, key=key)}


def _read_image(path: PathLike) -> tuple[Path, dict[str, ArrayImageReader]]:
    """Read image."""
    from PIL import Image
    from skimage.io import imread

    # disable decompression bomb protection
    Image.MAX_IMAGE_PIXELS = 30_000 * 30_000  # 30k x 30k pixels

    path = Path(path)
    assert path.exists(), f"File does not exist: {path}"
    key = get_key(path)
    return path, {path.name: ArrayImageReader(path, imread(path), key=key)}


def _read_npy_coordinates(path: PathLike) -> tuple[Path, dict[str, CoordinateImageReader]]:
    """Read data from npy file."""
    from image2image_io.readers.coordinate_reader import CoordinateImageReader

    path = Path(path)
    with open(path, "rb") as f:
        image = np.load(f)
    assert image.ndim == 2, "Only 2D images are supported"
    y, x = get_yx_coordinates_from_shape(image.shape)
    key = get_key(path)
    return path, {path.name: CoordinateImageReader(path, x, y, array_or_reader=image, key=key)}


def _read_npz_coordinates(path: PathLike) -> tuple[Path, dict[str, CoordinateImageReader]]:
    """Read data from npy file."""
    from image2image_io.readers.array_reader import ArrayImageReader

    path = Path(path)
    with np.load(path) as f:
        if "labels" in f:
            labels = f["labels"]
        elif "ppm_labels" in f:
            labels = f["ppm_labels"]
        else:
            raise ValueError("Labels not found in the file")
        labels = labels.tolist()
        if "array" in f:
            image = f["array"]
        elif "image" in f:
            image = f["image"]
        else:
            raise ValueError("Image not found in the file")
    assert image.ndim == 3, "Only 3D images are supported"
    key = get_key(path)
    return path, {path.name: ArrayImageReader(path, array=image, key=key, channel_names=labels)}


def _get_resolution_from_metadata(path: PathLike) -> float:
    """Get resolution from metadata."""
    from koyo.json import read_json_data

    resolution = 1.0
    path = Path(path)
    if path.name != "metadata.json":
        path_ = path / "metadata.json"
        if not path_.exists():
            path_ = path.parent / "metadata.json"
        path = path_
    if path.exists():
        metadata = read_json_data(path)
        resolution = metadata["metadata.experimental"]["pixel_size"]
    return resolution


def _read_metadata_h5_coordinates(path: PathLike) -> tuple[Path, dict[str, CoordinateImageReader]]:
    """Read coordinates from HDF5 file."""
    import h5py

    from image2image_io.readers.coordinate_reader import CoordinateImageReader

    path = Path(path)
    assert path.suffix in H5_EXTENSIONS, "Only .h5 files are supported"

    # read coordinates
    with h5py.File(path, "r") as f:
        try:
            yx = f["Dataset/Spatial/Coordinates/yx"][:]
            tic = f["Dataset/Spatial/Sum/y"][:]
        except KeyError:
            yx = f["Dataset/Spectral/Coordinates/yx"][:]
            tic = f["Dataset/Spectral/Sum/y"][:]

    y = yx[:, 0]
    x = yx[:, 1]
    resolution = _get_resolution_from_metadata(path)
    key = get_key(path)
    return path, {
        path.name: CoordinateImageReader(path, x, y, resolution=resolution, array_or_reader=reshape(x, y, tic), key=key)
    }


def _read_centroids_h5_coordinates(path: PathLike) -> tuple[Path, dict[str, CoordinateImageReader]]:
    """Read centroids data from HDF5 file."""
    path = Path(path)
    assert path.suffix in H5_EXTENSIONS, "Only .h5 files are supported"

    metadata_file = path.parent / "dataset.metadata.h5"
    if not metadata_file.exists():
        data_dir = path.parent.parent.with_suffix(".data")
        metadata_file = data_dir / "dataset.metadata.h5"
    if metadata_file.exists():
        return _read_centroids_h5_coordinates_with_metadata(path, metadata_file)
    return _read_centroids_h5_coordinates_without_metadata(path)


def _is_h5_mask(path: PathLike) -> bool:
    """Check whether this is a mask file."""
    import h5py

    path = Path(path)
    with h5py.File(path, "r") as f:
        return "Mask" in f


def _read_mask_h5(path: PathLike) -> tuple[Path, dict[str, ArrayImageReader]]:
    """Read centroids data from HDF5 file."""
    import h5py

    path = Path(path)
    assert path.suffix in H5_EXTENSIONS, "Only .h5 files are supported"

    # read coordinates
    with h5py.File(path, "r") as f:
        mask = f["Mask/mask"][:]
    resolution = 1.0
    if path.parent.name == "Masks":
        resolution = _get_resolution_from_metadata(path.parent.parent.with_suffix(".data"))
    key = get_key(path)
    return path, {path.name: ArrayImageReader(path, mask, resolution=resolution, key=key)}


def _read_centroids_h5_coordinates_lazy(path: PathLike) -> tuple[Path, dict[str, LazyCoordinateImageReader]]:
    """Read centroids data from HDF5 file."""
    path = Path(path)
    assert path.suffix in H5_EXTENSIONS, "Only .h5 files are supported"

    metadata_file = path.parent / "dataset.metadata.h5"
    if not metadata_file.exists():
        data_dir = path.parent.parent.with_suffix(".data")
        metadata_file = data_dir / "dataset.metadata.h5"
    if metadata_file.exists():
        return _read_centroids_h5_coordinates_with_metadata_lazy(path, metadata_file)
    return _read_centroids_h5_coordinates_without_metadata_lazy(path)


def _read_centroids_h5_coordinates_with_metadata(
    path: Path, metadata_file: Path
) -> tuple[Path, dict[str, CoordinateImageReader]]:
    import h5py

    from image2image_io.utils.utilities import format_mz

    assert metadata_file.exists(), f"File does not exist: {metadata_file}"
    _, reader_ = _read_metadata_h5_coordinates(metadata_file)
    key = next(iter(reader_.keys()))
    reader = reader_[key]

    x = reader.x
    y = reader.y

    with h5py.File(path, "r") as f:
        if "xs" in f["Array"]:
            mzs = f["Array"]["xs"][:]  # retrieve m/zs
            labels = [format_mz(mz) for mz in mzs]  # generate labels
        else:
            labels = f["Annotations/annotations/annotations"][:]
            labels = [label.decode() for label in labels]
        centroids = f["Array"]["array"][:]  # retrieve ion images
    centroids = reshape_batch(x, y, centroids)  # reshape images
    reader.data.update(dict(zip(labels, centroids)))
    return path, {path.name: reader}


def _read_centroids_h5_coordinates_with_metadata_lazy(
    path: Path, metadata_file: Path
) -> tuple[Path, dict[str, LazyCoordinateImageReader]]:
    import h5py

    from image2image_io.utils.lazy import LazyImageWrapper
    from image2image_io.utils.utilities import format_mz

    assert metadata_file.exists(), f"File does not exist: {metadata_file}"
    _, readers = _read_metadata_h5_coordinates(metadata_file)
    key = next(iter(readers.keys()))
    reader_ = readers[key]

    x = reader_.x
    y = reader_.y
    resolution = reader_.resolution

    with h5py.File(path, "r") as f:
        if "xs" in f["Array"]:
            mzs = f["Array"]["xs"][:]  # retrieve m/zs
            labels = [format_mz(mz) for mz in mzs]  # generate labels
        else:
            labels = f["Annotations/annotations/annotations"][:]
            labels = [label.decode() for label in labels]
    lazy_wrapper = LazyImageWrapper(path, "Array/array", labels, x, y)
    key = get_key(path)
    reader = LazyCoordinateImageReader(
        path, x, y, resolution=resolution, lazy_wrapper=lazy_wrapper, key=key, channel_names=labels
    )
    return path, {path.name: reader}


def _read_centroids_h5_coordinates_without_metadata(path: Path) -> tuple[Path, dict[str, CoordinateImageReader]]:
    import h5py

    from image2image_io.readers.coordinate_reader import CoordinateImageReader
    from image2image_io.utils.utilities import format_mz

    with h5py.File(path, "r") as f:
        # get coordinate metadata
        x = f["Misc/Spatial/x_coordinates"][:]
        y = f["Misc/Spatial/y_coordinates"][:]
        resolution = float(f["Misc/Spatial"].attrs["pixel_size"])
        if "xs" in f["Array"]:
            mzs = f["Array"]["xs"][:]  # retrieve m/zs
            labels = [format_mz(mz) for mz in mzs]  # generate labels
        else:
            labels = f["Annotations/annotations/annotations"][:]
            labels = [label.decode() for label in labels]
        centroids = f["Array"]["array"][:]  # retrieve ion images
        tic = np.random.randint(128, 255, len(x), dtype=np.uint8)
    tic = reshape(x, y, tic)
    key = get_key(path)
    reader = CoordinateImageReader(path, x, y, resolution=resolution, array_or_reader=tic, key=key)
    centroids = reshape_batch(x, y, centroids)  # reshape images
    reader.data.update(dict(zip(labels, centroids)))
    return path, {path.name: reader}


def _read_centroids_h5_coordinates_without_metadata_lazy(
    path: Path,
) -> tuple[Path, dict[str, LazyCoordinateImageReader]]:
    import h5py

    from image2image_io.readers.coordinate_reader import LazyCoordinateImageReader
    from image2image_io.utils.lazy import LazyImageWrapper
    from image2image_io.utils.utilities import format_mz

    with h5py.File(path, "r") as f:
        # get coordinate metadata
        x = f["Misc/Spatial/x_coordinates"][:]
        y = f["Misc/Spatial/y_coordinates"][:]
        resolution = float(f["Misc/Spatial"].attrs["pixel_size"])
        if "xs" in f["Array"]:
            mzs = f["Array"]["xs"][:]  # retrieve m/zs
            labels = [format_mz(mz) for mz in mzs]  # generate labels
        else:
            labels = f["Annotations/annotations/annotations"][:]
            labels = [label.decode() for label in labels]
    lazy_wrapper = LazyImageWrapper(path, "Array/array", labels, x, y)
    key = get_key(path)
    reader = LazyCoordinateImageReader(
        path, x, y, resolution=resolution, lazy_wrapper=lazy_wrapper, key=key, channel_names=labels
    )
    return path, {path.name: reader}


def _read_tsf_tdf_coordinates(
    path: PathLike,
    split_roi: bool = True,
    include_all: bool = True,
    scene_index: int | None = None,
) -> tuple[Path, dict[str, CoordinateImageReader]]:
    """Read coordinates from TSF file."""
    import sqlite3

    from image2image_io.readers.coordinate_reader import CoordinateImageReader

    path = Path(path)
    assert path.suffix in BRUKER_EXTENSIONS, "Only .tsf and .tdf files are supported"

    if path.suffix == ".d":
        if (path / "analysis.tsf").exists():
            path = path / "analysis.tsf"
        else:
            path = path / "analysis.tdf"

    # get wrapper
    conn = sqlite3.connect(path)

    try:
        cursor = conn.execute("SELECT SpotSize FROM MaldiFrameLaserInfo")
        resolution = float(cursor.fetchone()[0])
    except sqlite3.OperationalError:
        resolution = 1.0

    # get coordinates
    cursor = conn.cursor()
    cursor.execute("SELECT Frame, RegionNumber, XIndexPos, YIndexPos FROM MaldiFrameInfo")
    frame_index_position = np.array(cursor.fetchall())

    # get tic
    cursor = conn.execute("SELECT SummedIntensities FROM Frames")
    tic = np.array(cursor.fetchall())
    tic = tic[:, 0]

    # generate reader(s)
    readers = {}
    roi = frame_index_position[:, 1]
    unq_roi = np.unique(roi)
    if len(unq_roi) == 1:
        split_roi = False
        scene_index = None
        include_all = True

    # check whether scene index is correct
    if scene_index is not None:
        if scene_index not in unq_roi:
            raise ValueError(f"Scene index {scene_index} not found in the file")
        include_all = False
        unq_roi = [scene_index]
        split_roi = True

    # if we are not on macOS and there is only 1 ROI, we might as well get the proper data reader
    if not IS_MAC and not split_roi:
        return _read_tsf_tdf_reader(path)

    if include_all or not split_roi:  # or len(unq_roi) == 1):
        x = frame_index_position[:, 2]
        x = x - np.min(x)  # minimized
        y = frame_index_position[:, 3]
        y = y - np.min(y)  # minimized
        key = get_key(path)
        readers[path.name] = CoordinateImageReader(
            path, x, y, resolution=resolution, array_or_reader=reshape(x, y, tic), key=key
        )

    if split_roi:
        for current_scene_index in unq_roi:
            mask = roi == current_scene_index
            x_ = frame_index_position[mask, 2]
            x_ = x_ - np.min(x_)
            y_ = frame_index_position[mask, 3]
            y_ = y_ - np.min(y_)
            tic_ = tic[mask]
            key = get_key(path, current_scene_index)
            readers[f"S{current_scene_index}_{path.name}"] = CoordinateImageReader(
                path,
                x_,
                y_,
                resolution=resolution,
                array_or_reader=reshape(x_, y_, tic_),
                key=key,
                reader_kws={"scene_index": current_scene_index},
            )
    return path.parent, readers


def _read_tsf_tdf_reader(path: PathLike) -> tuple[Path, dict[str, CoordinateImageReader]]:
    """Read coordinates from Bruker file."""
    import sqlite3

    from imzy import get_reader

    from image2image_io.readers.coordinate_reader import CoordinateImageReader

    path = Path(path)
    assert path.suffix in BRUKER_EXTENSIONS, "Only .tsf and .tdf files are supported"

    if path.suffix == ".d":
        if (path / "analysis.tsf").exists():
            path = path / "analysis.tsf"
        else:
            path = path / "analysis.tdf"

    conn = sqlite3.connect(path)
    try:
        cursor = conn.execute("SELECT SpotSize FROM MaldiFrameLaserInfo")
        resolution = float(cursor.fetchone()[0])
    except sqlite3.OperationalError:
        resolution = 1.0
    conn.close()

    # get wrapper
    path = path.parent
    reader = get_reader(path)
    x = reader.x_coordinates
    y = reader.y_coordinates
    key = get_key(path)
    return path, {path.name: CoordinateImageReader(path, x, y, resolution=resolution, array_or_reader=reader, key=key)}


def _read_imzml_coordinates(path: PathLike) -> tuple[Path, dict[str, CoordinateImageReader]]:
    """Read coordinates from imzML file."""
    from imzy import get_reader

    from image2image_io.readers.coordinate_reader import CoordinateImageReader

    path = Path(path)
    assert path.suffix.lower() in IMZML_EXTENSIONS, "Only .imzML files are supported"
    if path.suffix.lower() == ".ibd":
        path = path.with_suffix(".imzML")

    # get wrapper
    reader = get_reader(path)
    x = reader.x_coordinates
    x = x - np.min(x)  # minimized
    y = reader.y_coordinates
    y = y - np.min(y)  # minimized
    tic = reader.get_tic()
    key = get_key(path)
    return path, {path.name: CoordinateImageReader(path, x, y, array_or_reader=reshape(x, y, tic), key=key)}


def _read_imzml_reader(path: PathLike) -> tuple[Path, dict[str, CoordinateImageReader]]:
    """Read coordinates from imzML file."""
    from imzy import get_reader

    from image2image_io.readers.coordinate_reader import CoordinateImageReader

    path = Path(path)
    assert path.suffix.lower() in IMZML_EXTENSIONS, "Only .imzML files are supported"
    if path.suffix.lower() == ".ibd":
        path = path.with_suffix(".imzML")

    # get wrapper
    reader = get_reader(path)
    x = reader.x_coordinates
    x = x - np.min(x)  # minimized
    y = reader.y_coordinates
    y = y - np.min(y)  # minimized
    key = get_key(path)
    return path, {path.name: CoordinateImageReader(path, x, y, array_or_reader=reader, key=key)}


def get_czi_metadata(path: PathLike) -> dict[str, dict[str, ty.Any]]:
    """Read CZI metadata."""
    from image2image_io.readers._czi import CziSceneFile

    path = Path(path)
    assert path.suffix == ".czi", "Only .czi files are supported"

    metadata = {}
    n = CziSceneFile.get_num_scenes(path)
    if n == 1:
        metadata[get_key(path)] = {"path": path, "scene_index": None}
    else:
        for scene_index in range(n):
            metadata[get_key(path, scene_index)] = {"path": path, "scene_index": scene_index}
    return metadata



"""Merge utility class."""

from __future__ import annotations

import typing as ty
from pathlib import Path
from warnings import warn

import numpy as np
from loguru import logger

if ty.TYPE_CHECKING:
    from image2image_io.readers import BaseReader


class MergeImages:
    """Merge multiple images."""

    def __init__(
        self,
        paths_or_readers: list[Path | str | BaseReader],
        pixel_sizes: list[int | float],
        channel_names: list[list[str]] | None = None,
        channel_ids: list[list[int]] | None = None,
        channel_colors: list[list[str]] | None = None,
    ):
        from image2image_io.readers import BaseReader, get_simple_reader

        if not isinstance(paths_or_readers, list):
            raise ValueError("MergeImages requires a list of images to merge")
        if not isinstance(pixel_sizes, list):
            raise ValueError("MergeImages requires a list of image resolutions for each image to merge")

        if channel_names is None:
            channel_names = [None] * len(paths_or_readers)
        if channel_ids is None:
            channel_ids = [None] * len(paths_or_readers)
        if channel_colors is None:
            channel_colors = [None] * len(paths_or_readers)

        readers = []
        for _index, (path, pixel_size, channel_names_, channel_ids_, channel_colors_) in enumerate(
            zip(paths_or_readers, pixel_sizes, channel_names, channel_ids, channel_colors)
        ):
            if isinstance(path, BaseReader):
                reader = path
            else:
                reader: BaseReader = get_simple_reader(path)
            if channel_colors_:
                reader._channel_colors = channel_colors_
            if channel_names_:
                reader._channel_names = channel_names_
            if channel_ids_:
                reader._channel_ids = channel_ids_
            reader.resolution = pixel_size
            if reader.channel_names is None:  # or len(reader.channel_names) != reader.n_channels:
                reader._channel_names = [f"C{idx}" for idx in range(0, reader.n_channels)]
                logger.trace(
                    f"Channel names not provided for {reader.path}, using {reader.channel_names} ({channel_names_})"
                )
            readers.append(reader)
        if not all(im.dtype == readers[0].dtype for im in readers):
            warn("MergeImages created with mixed data types, writing will cast to the largest data type", stacklevel=2)
        if any(im.is_rgb for im in readers):
            warn(
                "MergeImages does not support writing merged interleaved RGB. Data will be written as multi-channel",
                stacklevel=2,
            )

        self.readers: list[BaseReader] = readers
        self.paths: list[str | Path | BaseReader] = paths_or_readers
        self.dtype: np.dtype = self.readers[0].dtype
        self.is_rgb: bool = False
        self.n_channels: int = np.sum([reader.n_channels for reader in self.readers])
        self.channel_names: list[list[str]] = [reader.channel_names for reader in self.readers]
        self.channel_ids: list[list[int]] = [reader.channel_ids for reader in self.readers]
        self.original_size_transform = None

    @property
    def channel_names_flat(self) -> list[str]:
        """Return flat channel names."""
        return [name for names in self.channel_names for name in names]



"""Main CLI."""

import click
import koyo.compat
from loguru import logger

from image2image_io import __version__
from image2image_io.cli.convert import convert
from image2image_io.cli.czi2tiff import czi2tiff
from image2image_io.cli.merge import merge
from image2image_io.cli.thumbnail import thumbnail
from image2image_io.cli.transform import transform

LOG_FMT = "[<level>{level: <8}</level>][{time:YYYY-MM-DD HH:mm:ss:SSS}][{extra[src]}] {message}"
COLOR_LOG_FMT = (
    "<green>[<level>{level: <8}</level>]</green>"
    "<cyan>[{time:YYYY-MM-DD HH:mm:ss:SSS}]</cyan>"
    "<red>[{process}]</red>"
    "<blue>[{extra[src]}]</blue>"
    " {message}"
)


@click.group(
    context_settings={
        "help_option_names": ["-h", "--help"],
        "max_content_width": 120,
        "ignore_unknown_options": True,
    }
)
@click.option(
    "--dev",
    help="Flat to indicate that CLI should run in development mode and catch all errors.",
    default=False,
    is_flag=True,
    show_default=True,
)
@click.option(
    "--no_color",
    help="Flag to enable colored while doing tasks.",
    default=False,
    is_flag=True,
    show_default=True,
)
@click.version_option(__version__)
@click.option(
    "--verbose",
    "-v",
    "verbosity",
    default=1,
    count=True,
    help="Verbose output. This is additive flag so `-vvv` will print `INFO` messages and -vvvv will print `DEBUG`"
    " information.",
)
@click.option("--quiet", "-q", "verbosity", flag_value=0, help="Minimal output")
@click.option("--debug", "verbosity", flag_value=0.5, help="Maximum output")
def cli(verbosity: int, no_color: bool, dev: bool) -> None:
    """image2image-io CLI."""
    from koyo.hooks import install_debugger_hook, uninstall_debugger_hook
    from koyo.logging import get_loguru_config, set_loguru_env, set_loguru_log

    verbosity = 2 - int(verbosity)  # default is INFO
    if verbosity < 0:
        verbosity = 0
    if dev:
        install_debugger_hook()
        verbosity = 0
    elif dev:
        uninstall_debugger_hook()
    level = verbosity * 10
    level, fmt, colorize, enqueue = get_loguru_config(level, no_color=no_color)
    set_loguru_env(fmt, level, colorize, enqueue)
    set_loguru_log(
        level=level.upper(),
        no_color=no_color,
        logger=logger,
        fmt=LOG_FMT if no_color else COLOR_LOG_FMT,
    )
    logger.configure(extra={"src": "CLI"})
    logger.enable("image2image_io")
    logger.enable("koyo")
    logger.debug(f"Activated logger with level '{level}'.")


# register commands
cli.add_command(czi2tiff)
cli.add_command(thumbnail)
cli.add_command(merge)
cli.add_command(convert)
if transform:
    cli.add_command(transform)


def main():
    """Execute the "imimspy" command line program."""
    cli.main(windows_expand_args=False)


if __name__ == "__main__":
    cli.main(windows_expand_args=False)



"""Test shapes reader and utilities."""

import pandas as pd
import pytest

from image2image_io.readers.shapes_reader import read_shapes_from_df


@pytest.mark.parametrize("columns", [["vertex_x", "vertex_y", "cell"], ["vertex_x", "vertex_y", "cell_id"]])
def test_read_points_from_df(columns):
    x_col, y_col, cell_col = columns
    df = pd.DataFrame({x_col: [1, 2, 3], y_col: [4, 5, 6], cell_col: ["a", "a", "a"]})
    shapes_geojson, shapes_data, is_points = read_shapes_from_df(df)
    assert len(shapes_data) == 1, "Expected to read shapes from DataFrame."
    assert not is_points, "Expected to read shapes from DataFrame."

    df = pd.DataFrame({x_col: [1, 2, 3], y_col: [4, 5, 6], cell_col: ["a", "b", "c"]})
    shapes_geojson, shapes_data, is_points = read_shapes_from_df(df)
    assert len(shapes_data) == 3, "Expected to read shapes from DataFrame."
    assert not is_points, "Expected to read shapes from DataFrame."



"""Configuration to override few parameters."""

import typing as ty
from contextlib import contextmanager

from koyo.config import BaseConfig
from loguru import logger
from pydantic import Field, field_serializer, field_validator

from image2image_io.enums import ViewType
from image2image_io.utils._appdirs import USER_CONFIG_DIR


# noinspection PyMethodParameters
class Config(BaseConfig):
    """Configuration of few parameters."""

    USER_CONFIG_DIR = USER_CONFIG_DIR
    USER_CONFIG_FILENAME = "config_reader.json"

    quiet: bool = Field(
        False,
        title="Quiet",
        description="Quiet mode.",
        json_schema_extra={
            "in_app": True,
        },
    )
    init_pyramid: bool = Field(
        True,
        title="Initialize pyramid",
        description="Initialize pyramid on data load.",
        json_schema_extra={
            "in_app": True,
        },
    )
    auto_pyramid: bool = Field(
        True,
        title="Auto pyramid",
        description="Automatically create pyramid.",
        json_schema_extra={
            "in_app": True,
        },
    )
    split_czi: bool = Field(
        True,
        title="Split CZI image by scene",
        description="When loading CZI image(s), split them by the scene",
        json_schema_extra={
            "in_app": True,
        },
    )
    split_rgb: bool = Field(
        True,
        title="Split RGB image by channel",
        description="When loading RGB image(s), split them by the channel",
        json_schema_extra={
            "in_app": True,
        },
    )
    split_roi: bool = Field(
        True,
        title="Split Bruker .d images by region of interest.",
        description="When loading Bruker .d image(s), split them by the region of interest.",
        json_schema_extra={
            "in_app": True,
        },
    )
    only_last_pyramid: bool = Field(
        False,
        title="Only last pyramid",
        description="Only use last pyramid.",
        json_schema_extra={
            "in_app": True,
        },
    )
    view_type: ViewType = Field(
        ViewType.OVERLAY,
        title="View type",
        description="IMS view type.",
        json_schema_extra={
            "in_app": False,
        },
    )
    show_transformed: bool = Field(
        True,
        title="Show transformed",
        description="If checked, transformed moving image will be shown.",
        json_schema_extra={
            "in_app": False,
        },
    )
    shape_display: ty.Literal["polygon", "path", "polygon or path", "points"] = Field(
        "path",
        title="Shape display",
        description="Shape display type.",
        json_schema_extra={
            "in_app": False,
        },
    )
    subsample: bool = Field(
        True,
        title="Subsample",
        description="Subsample shapes for display.",
        json_schema_extra={
            "in_app": False,
        },
    )
    subsample_ratio: float = Field(
        0.01,
        title="Subsample ratio",
        description="Shapes subsample ratio.",
        json_schema_extra={
            "in_app": False,
        },
    )
    subsample_random_seed: int = Field(
        -1,
        title="Subsample random seed",
        description="Shapes subsample ratio.",
        json_schema_extra={
            "in_app": False,
            "save": False,
        },
    )
    multicore: bool = Field(
        True,
        title="Multicore",
        description="Use multicore processing.",
        json_schema_extra={
            "in_app": False,
        },
    )

    @field_validator("view_type", mode="before")
    @classmethod
    def _validate_view_type(cls, value: ty.Union[str, ViewType]) -> ViewType:  # type: ignore[misc]
        """Validate view_type."""
        return ViewType(value)

    @field_serializer("view_type", when_used="always")
    def _serialize_view_type(self, value: ViewType) -> str:
        """Serialize view_type."""
        if hasattr(value, "value"):
            return value.value
        return str(value)

    @contextmanager
    def temporary_override(self, **kwargs: ty.Any) -> ty.Generator[None, None, None]:
        """Temporary override configuration."""
        old_values = {key: getattr(self, key) for key in kwargs}
        for key, value in kwargs.items():
            setattr(self, key, value)
        yield
        for key, value in old_values.items():
            setattr(self, key, value)

    def trace(self, msg: str) -> None:
        """Trace logging."""
        if not self.quiet:
            logger.trace(msg)


CONFIG = Config()  # type: ignore[call-arg]


def get_reader_config() -> Config:
    """Get reader configuration."""
    global CONFIG

    if CONFIG is None:
        CONFIG = Config()
        CONFIG.load()
    return CONFIG



"""Utilities."""

from __future__ import annotations

import click
from koyo.click import Parameter, cli_parse_paths_sort, print_parameters

from image2image_io.cli._common import ALLOW_EXTRA_ARGS


@click.option(
    "-W",
    "--overwrite",
    help="Overwrite existing data.",
    is_flag=True,
    default=None,
    show_default=True,
)
@click.option(
    "-u/-U",
    "--as_uint8/--no_as_uint8",
    help="Downcast the image data format to uint8 which will substantially reduce the size of the files (unless it's"
    " already in uint8...).",
    is_flag=True,
    default=None,
    show_default=True,
)
@click.option(
    "-t",
    "--tile_size",
    help="Tile size.",
    type=click.Choice(["256", "512", "1024", "2048"], case_sensitive=False),
    default="512",
    show_default=True,
    required=False,
)
@click.option(
    "-f",
    "--fmt",
    help="Output format.",
    type=click.Choice(["ome-tiff"], case_sensitive=False),
    default="ome-tiff",
    show_default=True,
    required=False,
)
@click.option(
    "-o",
    "--output_dir",
    help="Path to the output directory where images should be saved to.",
    type=click.Path(file_okay=False, dir_okay=True, resolve_path=True),
    show_default=True,
    required=True,
)
@click.option(
    "-i",
    "--image",
    help="Path(s) of images to be converted.",
    type=click.UNPROCESSED,
    show_default=True,
    required=True,
    multiple=True,
    callback=cli_parse_paths_sort,
)
@click.command("convert", context_settings=ALLOW_EXTRA_ARGS)
def convert(image: list[str], output_dir: str, fmt: str, tile_size: str, as_uint8: bool, overwrite: bool) -> None:
    """Convert images to pyramidal OME-TIFF."""
    convert_runner(image, output_dir, fmt, tile_size, as_uint8, overwrite)


def convert_runner(
    paths: list[str],
    output_dir: str,
    fmt: str = "ome-tiff",
    tile_size: int | str = 512,
    as_uint8: bool | None = None,
    overwrite: bool = False,
) -> None:
    """Convert images to pyramidal OME-TIFF."""
    from image2image_io.writers import images_to_ome_tiff

    print_parameters(
        Parameter("Images", "-i/--image", paths),
        Parameter("Output directory", "-o/--output_dir", output_dir),
        Parameter("Output format", "-f/--fmt", fmt),
        Parameter("Tile size", "-t/--tile_size", tile_size),
        Parameter("Downcast to uint8", "-u/--as_uint8", as_uint8),
        Parameter("Overwrite", "-W/--overwrite", overwrite),
    )
    for _ in images_to_ome_tiff(paths, output_dir, tile_size=int(tile_size), as_uint8=as_uint8, overwrite=overwrite):
        pass
